#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`695` **utility data submodule.**

This submodule exercises :pep:`695` support for statement-level type aliases
implemented in the :func:`beartype.beartype` decorator by declaring unit tests
exercising these aliases. For safety, these tests are intentionally isolated
from the main test suite. Notably, this low-level submodule implements the
higher-level ``test_decor_pep695()`` unit test in the main test suite.

Caveats
-------
**This submodule requires the active Python interpreter to target at least
Python 3.12.0.** If this is *not* the case, importing this submodule raises an
:exc:`SyntaxError` exception.
'''

# ....................{ TESTS ~ tester                     }....................
def unit_test_is_hint_pep695_subbed() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep695.is_hint_pep695_subbed`
    tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.hint.pep.proposal.pep695 import (
        is_hint_pep695_subbed)

    # ....................{ LOCALS                         }....................
    # Unsubscripted type alias.
    type and_thou = int | float

    # Subscriptable type alias yet to be subscripted by a concrete type.
    #
    # Note that, although subscriptable type aliases superficially appear to be
    # "pre-subscripted" by PEP 484-compliant type variables, this
    # "pre-subscription" is simply syntactic sugar; subscriptable type aliases
    # remain unsubscripted until explicitly subscripted by concrete types.
    type colossal_skeleton[T] = complex | T

    # ....................{ PASS                           }....................
    # Assert this tester accepts PEP 695-compliant subscripted type aliases
    # (i.e., subscriptable type aliases subscripted by concrete types).
    assert is_hint_pep695_subbed(colossal_skeleton[int]) is True

    # ....................{ FAIL                           }....................
    # Assert this tester rejects objects that are *NOT* PEP 585-compliant
    # subscripted builtins.
    assert is_hint_pep695_subbed(
        'And thou, colossal Skeleton, that, still') is False

    # Assert this tester rejects PEP 585-compliant subscripted builtins that are
    # *NOT* PEP 695-compliant subscripted type aliases.
    assert is_hint_pep695_subbed(list[str]) is False

    # Assert this tester rejects PEP 695-compliant unsubscripted type aliases.
    assert is_hint_pep695_subbed(and_thou) is False

    # Assert this tester accepts PEP 695-compliant subscriptable type aliases
    # yet to be subscripted by a concrete type.
    assert is_hint_pep695_subbed(colossal_skeleton) is False

# ....................{ TESTS ~ getter                     }....................
def unit_test_get_hint_pep695_parameterizable_typeparams() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep695._get_hint_pep695_parameterizable_typeparams`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.typing import (
        ParamSpec,
        TypeVar,
        TypeVarTuple,
    )
    from beartype._util.hint.pep.proposal.pep695 import (
        _get_hint_pep695_parameterizable_typeparams)

    # ....................{ CLASSES                        }....................
    class AndThenSpake[S]:  # <-- this is madness. this is PEP 695.
        '''
        Class parametrized by *only* a :pep:`484`-compliant type variable.

        Note that this class intentionally avoids subclassing *any* superclass,
        including:

        * The root :class:`object` superclass, as doing so raises unreadable
          low-level Python exceptions resembling:

          TypeError: Cannot create a consistent method resolution order (MRO)
          for bases object, Generic

        * The :pep:`484`-compliant :class:`typing.Generic` superclass, as doing
          so raises unreadable low-level Python exceptions resembling:

          TypeError: Cannot inherit from Generic[...] multiple times.

        This is nonsensical and actively obstructs the declaration of generics.
        Explicit is better than implicit -- especially when implicit stopped
        making sense 50 PEP standards and 4 CPython minor versions ago.
        '''

        pass

    # ....................{ CALLABLES                      }....................
    def as_with_a_palsied_tongue[*Ss, **Q]():
        '''
        Function parametrized by *only* a :pep:`612`-compliant parameter
        specifications and :pep:`646`-compliant tuple type variable.
        '''

        pass

    # ....................{ LOCALS                         }....................
    # Type alias parametrized by one of each possible type parameter.
    type and_that_fair_kneeling_goddess[T, *Ts, **P] = int

    # ....................{ ASSERTS                        }....................
    # Assert this getter passed a pure-Python class returns the expected tuple.
    AndThenSpake_typevars = _get_hint_pep695_parameterizable_typeparams(
        AndThenSpake)
    assert len(AndThenSpake_typevars) == 1
    assert isinstance(AndThenSpake_typevars[0], TypeVar)

    # Assert this getter passed a pure-Python function returns the expected
    # tuple.
    as_with_a_palsied_tongue_typevars = (
        _get_hint_pep695_parameterizable_typeparams(as_with_a_palsied_tongue))
    assert len(as_with_a_palsied_tongue_typevars) == 2
    assert isinstance(as_with_a_palsied_tongue_typevars[0], TypeVarTuple)
    assert isinstance(as_with_a_palsied_tongue_typevars[1], ParamSpec)

    # Assert this getter passed a PEP 695-compliant type alias returns the
    # expected tuple.
    and_that_fair_kneeling_goddess_typevars = (
        _get_hint_pep695_parameterizable_typeparams(
            and_that_fair_kneeling_goddess))
    assert len(and_that_fair_kneeling_goddess_typevars) == 3
    assert isinstance(and_that_fair_kneeling_goddess_typevars[0], TypeVar)
    assert isinstance(and_that_fair_kneeling_goddess_typevars[1], TypeVarTuple)
    assert isinstance(and_that_fair_kneeling_goddess_typevars[2], ParamSpec)

# ....................{ TESTS ~ iterator                   }....................
def unit_test_iter_hint_pep695_forwardrefs() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep695.iter_hint_pep695_unsubbed_forwardrefs`
    iterator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep695Exception
    from beartype._check.forward.reference.fwdrefmeta import (
        BeartypeForwardRefMeta)
    from beartype._util.hint.pep.proposal.pep695 import (
        iter_hint_pep695_unsubbed_forwardrefs)
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # Type alias containing *NO* unquoted forward references.
    type of_intermitted_song = str | int

    # Type alias containing one unquoted forward reference.
    type sudden_she_rose = ItsBurstingBurthen | bool

    # Type alias containing two or more unquoted forward reference.
    type as_if_her_heart = ImpatientlyEndured | complex | AtTheSoundHeTurned

    # ....................{ ASSERTS ~ null                 }....................
    # Assert that this iterator yields nothing when passed a type alias
    # containing *NO* unquoted forward references.
    their_own_life = list(
        iter_hint_pep695_unsubbed_forwardrefs(of_intermitted_song))
    assert not their_own_life

    # ....................{ ASSERTS ~ single               }....................
    # Assert that this iterator yields a single forward reference proxy
    # referring to the single unquoted forward reference embedded in a
    # passed type alias containing only that reference.
    and_saw_by = iter_hint_pep695_unsubbed_forwardrefs(sudden_she_rose)
    the_warm_light_of = next(and_saw_by)
    assert isinstance(the_warm_light_of, BeartypeForwardRefMeta)
    assert the_warm_light_of.__name__ == 'ItsBurstingBurthen'

    # Assert this iterator raises the expected exception when attempting to
    # erroneously iterate past an unquoted forward reference referring to an
    # undefined attribute.
    with raises(BeartypeDecorHintPep695Exception):
        next(and_saw_by)

    class ItsBurstingBurthen(object):
        '''
        Class to which this unquoted forward reference refers, intentionally
        declared while iterating this iterator.
        '''

        pass

    # Assert that this iterator is now exhausted.
    with raises(StopIteration):
        next(and_saw_by)

    # ....................{ ASSERTS ~ multiple             }....................
    # Assert that this iterator first yields a forward reference proxy referring
    # to the first unquoted forward reference embedded in a passed type alias
    # containing only that reference.
    her_glowing_limbs = iter_hint_pep695_unsubbed_forwardrefs(
        as_if_her_heart)
    beneath_the_sinuous_veil = next(her_glowing_limbs)
    assert isinstance(beneath_the_sinuous_veil, BeartypeForwardRefMeta)
    assert beneath_the_sinuous_veil.__name__ == 'ImpatientlyEndured'

    class ImpatientlyEndured(object):
        '''
        Class to which this unquoted forward reference refers, intentionally
        declared while iterating this iterator.
        '''

        pass

    # Assert that this iterator lastly yields a forward reference proxy
    # referring to the last unquoted forward reference in this type alias.
    of_woven_wind = next(her_glowing_limbs)
    assert isinstance(of_woven_wind, BeartypeForwardRefMeta)
    assert of_woven_wind.__name__ == 'AtTheSoundHeTurned'

    class AtTheSoundHeTurned(object):
        '''
        Class to which this unquoted forward reference refers, intentionally
        declared while iterating this iterator.
        '''

        pass

    # Assert this iterator raises the expected exception when attempting to
    # erroneously iterate past an unquoted forward reference referring to an
    # undefined attribute.
    with raises(StopIteration):
        next(her_glowing_limbs)

# ....................{ TESTS ~ reducer                    }....................
def unit_test_reduce_hint_pep695_unsubbed() -> None:
    '''
    Test the public
    :mod:`beartype._check.convert._reduce._pep.redpep695.reduce_hint_pep695_unsubbed`
    reducer.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep695Exception
    from beartype._check.convert._reduce._pep.redpep695 import (
        reduce_hint_pep695_unsubbed)
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # Type alias containing *NO* unquoted forward references.
    type her_dark_locks = str | int

    # Type alias containing one unquoted forward reference.
    type floating_in = TheBreathOfNight | bool

    # ....................{ PASS                           }....................
    # Assert that this reducer returns the type hint underlying the passed type
    # alias containing *NO* unquoted forward references.
    hint_sane = reduce_hint_pep695_unsubbed(
        hint=her_dark_locks,
        hint_parent_sane=None,
        exception_prefix='',
    )
    assert hint_sane.hint == str | int

    # ....................{ FAIL                           }....................
    # Assert that this reducer raises the expected exception when passed a type
    # alias containing one unquoted forward reference.
    with raises(BeartypeDecorHintPep695Exception):
        reduce_hint_pep695_unsubbed(
            hint=floating_in,
            hint_parent_sane=None,
            exception_prefix='',
        )


def unit_test_reduce_hint_pep484612646_subbed_typeargs_to_hints_for_pep695() -> (
    None):
    '''
    Test the public ``reduce_hint_pep484612646_subbed_typeargs_to_hints()`` getter
    with respect to :pep:`695`-compliant subscripted type aliases.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep484612646Exception
    from beartype._check.convert._reduce._pep.redpep484612646 import (
        reduce_hint_pep484612646_subbed_typeargs_to_hints)
    from beartype._check.metadata.hint.hintsane import HintSane
    from beartype._util.hint.pep.utilpepget import get_hint_pep_typeargs_packed
    from beartype._util.kind.maplike.utilmapfrozen import FrozenDict
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # Unsubscripted type alias.
    type guiding_its = int | float

    # Subscriptable type alias correctly parametrized by a single type variable.
    type irresistible_career[T] = complex | T

    # Subscriptable type alias correctly parametrized by multiple type
    # variables.
    type in_thy[S, T] = list[S] | dict[S, T]

    # Tuples of all type variables parametrizing these type aliases.
    irresistible_career_typevars = get_hint_pep_typeargs_packed(irresistible_career)
    in_thy_typevars = get_hint_pep_typeargs_packed(in_thy)

    # ....................{ PASS                           }....................
    # Assert this getter passed PEP 695-compliant subscripted type aliases
    # returns the expected tuple of all type variables parametrizing the
    # unsubscripted type aliases underlying these subscripted type aliases.
    #
    # Note that these type variables are literally scoped (i.e., isolated) to
    # these aliases and thus accessible *ONLY* by directly accessing the
    # "__parameters__" dunder attribute on these aliases. It is what it is.
    hint_sane = reduce_hint_pep484612646_subbed_typeargs_to_hints(
        irresistible_career[int])
    assert hint_sane == HintSane(
        hint=irresistible_career,
        typearg_to_hint=FrozenDict({irresistible_career_typevars[0]: int,}),
    )

    hint_sane = reduce_hint_pep484612646_subbed_typeargs_to_hints(
        in_thy[bool, complex])
    assert hint_sane == HintSane(
        hint=in_thy,
        typearg_to_hint=FrozenDict({
            in_thy_typevars[0]: bool,
            in_thy_typevars[1]: complex,
        }),
    )

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when passed an object
    # that is *NOT* a PEP 695-compliant subscripted type alias.
    with raises(BeartypeDecorHintPep484612646Exception):
        reduce_hint_pep484612646_subbed_typeargs_to_hints(
            'In thy devastating omnipotence,')

    # Assert this getter raises the expected exception when passed a PEP
    # 695-compliant unsubscripted type alias.
    with raises(BeartypeDecorHintPep484612646Exception):
        reduce_hint_pep484612646_subbed_typeargs_to_hints(guiding_its)
    with raises(BeartypeDecorHintPep484612646Exception):
        reduce_hint_pep484612646_subbed_typeargs_to_hints(irresistible_career)

# ....................{ TESTS ~ decorator                  }....................
#FIXME: Actually call this unit test!
def unit_test_decor_hint_pep695_type_recursive_indirect() -> None:
    '''
    Test :pep:`695`-compliant type aliases with respect to a
    :func:`beartype.beartype`-decorated class parametrized by a
    :pep:`484`-compliant type variable bound to an unsubscripted type alias such
    that that class defines a method accepting a parameter annotated as that
    same type alias subscripted by that same type variable, indirectly inducing
    recursion between that type alias and type variable.

    This unit test exercises a non-trivial `edge case`_ that actually arose.

    .. _edge case:
       https://github.com/beartype/beartype/issues/510
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # Type alias aliasing a PEP 585-compliant fixed-length tuple type hint
    # subscripted by one type variable and an arbitrary builtin type.
    type WithStr[T] = tuple[T, str]

    # ....................{ CALLABLES                      }....................
    @beartype
    class ToOverbear[T: WithStr]:
        '''
        Arbitrary :func:`beartype.beartype`-decorated class parametrized by a
        :pep:`484`-compliant type variable bound to an unsubscripted type alias.
        '''

        @staticmethod
        def and_crumble_this(to_naught: WithStr[T]) -> str:
            '''
            Arbitrary method accepting a parameter annotated as that same type
            alias subscripted by that same type variable, indirectly inducing
            recursion between that type alias and type variable.
            '''

            (_, inner_item_second), outer_item_second = to_naught
            return inner_item_second + outer_item_second

    # ....................{ LOCALS                         }....................
    # Arbitrary tuple satisfying the "to_naught" parameter defined above with
    # respect to O(1) type-checking generated by @beartype.
    #
    # Note that this tuple is intentionally *NOT* infinitely recursive. For
    # efficiency, O(1) type-checking generated by @beartype only type-checks
    # PEP 695-compliant recursive type aliases to one layer of nesting. Ergo, a
    # tuple of tuple of strings suffices to exercise this functionality.
    where_is_another_chaos = (
        (
            # Technically, this should be an infinitely recursive tuple
            # exhibiting the same internal structure to satisfy this hint.
            # Pragmatically, @beartype only type-checks PEP 695-compliant
            # recursive type aliases to one layer of nesting as detailed above.
            # Since this is that layer of nesting, *ANY* object suffices.
            0xFEEDFACE,
            'Where is another chaos? Where?"—That word',
        ),
        'Found way unto Olympus, and made quake',
    )

    # Arbitrary tuple violating the "to_naught" parameter defined above with
    # respect to O(1) type-checking generated by @beartype.
    found_way_unto_olympus = (
        (
            0xCAFEBABE,  # See above.
            b'The rebel three.--Thea was startled up,',
        ),
        'Found way unto Olympus, and made quake',
    )

    # ....................{ PASS                           }....................
    # Assert that this method passed a valid parameter returns the expected
    # value.
    assert ToOverbear.and_crumble_this(where_is_another_chaos) == (
        'Where is another chaos? Where?"—That word'
        'Found way unto Olympus, and made quake'
    )

    # ....................{ FAIL                           }....................
    # Assert that this method passed an valid parameter raises the expected
    # exception.
    with raises(BeartypeCallHintParamViolation):
        ToOverbear.and_crumble_this(found_way_unto_olympus)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :mod:`contextlib` utilities (i.e., low-level callables handling the
standard :mod:`contextlib` module).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Any,
    Optional,
)
from beartype._util.func.utilfunccodeobj import (
    get_func_codeobj_or_none,
    get_func_codeobj_basename,
)
from beartype._util.py.utilpyversion import IS_PYTHON_AT_MOST_3_10
from collections.abc import (
    Callable,
    # Generator,
)

# ....................{ GETTERS                            }....................
#FIXME: Generalize into a new get_func_contextlib_contextmanager_or_none()
#function behaving as follows:
#* If the passed callable is decorated by @contextlib.contextmanager, return
#  the contextlib.contextmanager() decorator function.
#* If the passed callable is decorated by @contextlib.asynccontextmanager, return
#  the contextlib.asynccontextmanager() decorator function.
#* Else, return "None".
#FIXME: Unit test us up, please.
def get_func_contextlib_contextmanager_or_none(func: Any) -> Optional[Callable]:
    '''
    :mod:`contextlib` decorator underlying the passed object if this object is a
    :mod:`contextlib`-based **isomorphic decorator closure** (i.e., closure both
    defined and returned by either the standard
    :func:`contextlib.contextmanager` or :func:`contextlib.asynccontextmanager`
    decorators where that closure isomorphically preserves both the number and
    types of all passed parameters and returns by accepting only a variadic
    positional argument and variadic keyword argument) *or* :data:`None`
    otherwise (i.e., if this object is *not* such a closure).

    Specifically, this getter returns either:

    * If the passed object was produced by a prior call to the
      :func:`contextlib.contextmanager` decorator, that decorator as a function.
    * If the passed object was produced by a prior call to the
      :func:`contextlib.asynccontextmanager` decorator, that decorator as a
      function.
    * Else, :data:`None`.

    This getter enables callers to detect when a user-defined callable has been
    decorated by a :mod:`contextlib` decorator and thus has a mismatch between
    the type hints annotating that decorated callable and the type of the object
    created and returned by that decorated callable.

    Caveats
    -------
    **This getter only supports Python >= 3.11.** Under Python <= 3.10, this
    getter erroneously returns **false negatives** (i.e., :data:`False` even
    when the passed callable is a valid :func:`contextlib.contextmanager`-based
    isomorphic decorator closure). Why? Because this getter detects these
    closures by internally testing the ``co_qualname`` instance variable on the
    code object of the passed callable, which only exists under Python >= 3.11.
    This is not the fault of :mod:`beartype` and *totally* outside our control.

    Parameters
    ----------
    func : object
        Object to be inspected.

    Returns
    -------
    Optional[Callable]
        Either:

        * If this object is a :func:`contextlib.asynccontextmanager`-based
          isomorphic decorator closure, :func:`contextlib.asynccontextmanager`.
        * If this object is a :func:`contextlib.contextmanager`-based isomorphic
          decorator closure, :func:`contextlib.contextmanager`.
        * Else, :data:`None`.

    See Also
    --------
    :obj:`beartype._data.func.datafunc.CONTEXTLIB_CONTEXTMANAGER_CO_NAME_QUALNAME`
        Further discussion.
    '''

    # Avoid circular import dependencies.
    from beartype._util.func.utilfunctest import is_func_closure

    # If either...
    if (
        # The active Python interpreter targets Python < 3.10 and thus fails to
        # define the "co_qualname" attribute on code objects required to
        # robustly implement this test *OR*...
        IS_PYTHON_AT_MOST_3_10 or
        # The passed callable is *NOT* a closure...
        not is_func_closure(func)
    ):
        # Then immediately return "None".
        return None
    # Else, that callable is a closure.

    # Code object underlying that callable as is (rather than possibly unwrapped
    # to another code object entirely) if that callable is pure-Python *OR*
    # "None" otherwise (i.e., if that callable is C-based).
    func_codeobj = get_func_codeobj_or_none(func)

    # If that callable is C-based, immediately return "None".
    if func_codeobj is None:
        return None
    # Else, that callable is pure-Python.

    # Defer heavyweight getter-specific imports with potential side effects --
    # notably, increased costs to space and time complexity.
    from beartype._data.api.standard.datacontextlib import (
        CONTEXTLIB_CONTEXTMANAGER_CODEOBJ_NAME_TO_DECORATOR)

    # Fully-qualified name of this code object.
    func_codeobj_name = get_func_codeobj_basename(func_codeobj)

    # Either:
    # * If the fully-qualified name of this code object is that of an isomorphic
    #   decorator closure created and returned by a standard "contextlib"
    #   decorator, that decorator.
    # * Else, "None".
    contextlib_decorator = (
        CONTEXTLIB_CONTEXTMANAGER_CODEOBJ_NAME_TO_DECORATOR.get(
            func_codeobj_name))

    # Return this decorator.
    #
    # Note that we *COULD* technically also explicitly test whether that
    # callable satisfies the is_func_wrapper_isomorphic() getter -- but that
    # there's no benefit and a minor efficiency cost to doing so.
    return contextlib_decorator  # type: ignore[return-value]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide **path factories** (i.e., callables creating and returning
:class:`pathlib.Path` instances encapsulating testing-specific paths).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ RELATIVIZERS                       }....................
def shell_quote(text: str) -> str:
    '''
    Shell-quote the passed string in a platform-specific manner.

    If the current platform is:

    * *Not* vanilla Windows (e.g., Linux, macOS), the returned string is
      guaranteed to be suitable for passing as an arbitrary positional argument
      to external commands.
    * Windows, the returned string is suitable for passing *only* to external
      commands parsing arguments in the same manner as the Microsoft C runtime.
      While *all* applications on POSIX-compliant systems are required to parse
      arguments in the same manner (i.e., according to Bourne shell lexing), no
      such standard applies to Windows applications. Shell quoting is therefore
      fragile under Windows -- like pretty much everything.

    Parameters
    ----------
    text : str
        String to be shell-quoted.

    Returns
    -------
    str
        This string shell-quoted.
    '''
    assert isinstance(text, str), f'{repr(text)} not string.'

    # Defer heavyweight imports.
    from beartype._util.os.utilostest import is_os_windows_vanilla

    # If the current platform is vanilla Windows (i.e., neither Cygwin *NOR* the
    # Windows Subsystem for Linux (WSL)), do *NOT* perform POSIX-compatible
    # quoting. Vanilla Windows is POSIX-incompatible and hence does *NOT* parse
    # command-line arguments according to POSIX standards. In particular,
    # Windows does *NOT* treat single-quoted arguments as single arguments but
    # rather as multiple shell words delimited by the raw literal `'`. This is
    # circumventable by calling an officially undocumented Windows-specific
    # function. (Awesome.)
    if is_os_windows_vanilla():
        import subprocess
        return subprocess.list2cmdline([text])
    # Else, perform POSIX-compatible quoting.
    else:
        import shlex
        return shlex.quote(text)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :mod:`functools` utilities (i.e., low-level callables handling
functionality defined by the standard :mod:`functools` module).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorWrappeeException
from beartype.roar._roarexc import _BeartypeUtilCallableException
from beartype.typing import (
    Any,
    Tuple,
)
from beartype._cave._cavefast import (
    CallableFunctoolsLruCacheType,
    CallableFunctoolsPartialType,
)
from beartype._data.typing.datatypingport import TypeIs
from beartype._data.typing.datatyping import (
    BeartypeableT,
    DictStrToAny,
    TypeException,
)
from collections.abc import Callable
from functools import lru_cache

# ....................{ TESTERS                            }....................
def is_func_functools_lru_cache(func: Any) -> TypeIs[Callable]:
    '''
    :data:`True` only if the passed object is a
    :func:`functools.lru_cache`-memoized **pseudo-callable** (i.e., low-level
    C-based callable object both created and returned by the standard
    :func:`functools.lru_cache` decorator).

    This tester enables callers to detect when a user-defined callable has been
    decorated by the :func:`functools.lru_cache` decorator, which creates
    low-level C-based callable objects requiring special handling elsewhere.

    Parameters
    ----------
    func : object
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a
        :func:`functools.lru_cache`-memoized callable.
    '''

    # Defer heavyweight tester-specific imports with potential side effects --
    # notably, increased costs to space and time complexity.

    # Return true only if the type of that callable is the low-level C-based
    # private type of all objects created and returned by the standard
    # @functools.lru_cache decorator.
    return isinstance(func, CallableFunctoolsLruCacheType)


def is_func_functools_partial(func: Any) -> TypeIs[
    CallableFunctoolsPartialType]:
    '''
    :data:`True` only if the passed object is a **partial** (i.e., pure-Python
    callable :class:`functools.partial` object wrapping a possibly C-based
    callable).

    Parameters
    ----------
    func : object
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a
        :func:`functools.partial`-wrapped callable.
    '''

    # Return true only if the type of that callable is the high-level
    # pure-Python public type of all objects created and returned by the
    # standard functools.partial() factory.
    return isinstance(func, CallableFunctoolsPartialType)

# ....................{ GETTERS                            }....................
def get_func_functools_partial_args(
    func: CallableFunctoolsPartialType) -> Tuple[tuple, DictStrToAny]:
    '''
    2-tuple ``(args, kwargs)`` providing the positional and keyword parameters
    with which the passed **partial** (i.e., pure-Python callable
    :class:`functools.partial` object directly wrapping this possibly C-based
    callable) was originally partialized.

    Parameters
    ----------
    func : CallableFunctoolsPartialType
        Partial to be inspected.

    Returns
    -------
    Tuple[tuple, DictStrToAny]
        2-tuple ``(args, kwargs)`` such that:

        * ``args`` is the tuple of the zero or more positional parameters passed
          to the callable partialized by this partial.
        * ``kwargs`` is the dictionary mapping from the name to value of the
          zero or more keyword parameters passed to the callable partialized by
          this partial.
    '''
    assert isinstance(func, CallableFunctoolsPartialType), (
        f'{repr(func)} not "function.partial"-wrapped callable.')

    # Return a 2-tuple providing the positional and keyword parameters with
    # which this partial was originally partialized.
    return (func.args, func.keywords)


def get_func_functools_partial_args_flexible_len(
    # Mandatory parameters.
    func: CallableFunctoolsPartialType,

    # Optional parameters.
    is_unwrap: bool = True,
    exception_cls: TypeException = _BeartypeUtilCallableException,
    exception_prefix: str = '',
) -> int:
    '''
    Number of **flexible parameters** (i.e., parameters passable as either
    positional or keyword arguments but *not* positional-only, keyword-only,
    variadic, or other more constrained kinds of parameters) accepted by the
    passed **partial** (i.e., pure-Python callable :class:`functools.partial`
    object directly wrapping this possibly C-based callable).

    Specifically, this getter transparently returns the total number of flexible
    parameters accepted by the lower-level callable wrapped by this partial
    minus the number of flexible parameters partialized away by this partial.

    Parameters
    ----------
    func : CallableFunctoolsPartialType
        Partial to be inspected.
    is_unwrap: bool, optional
        :data:`True` only if this getter implicitly calls the
        :func:`beartype._util.func.utilfuncwrap.unwrap_func_all` function.
        Defaults to :data:`True` for safety. See :func:`.get_func_codeobj` for
        further commentary.
    exception_cls : type, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :class:`._BeartypeUtilCallableException`.
    exception_prefix : str, optional
        Human-readable label prefixing the message of any exception raised in
        the event of a fatal error. Defaults to the empty string.

    Returns
    -------
    int
        Number of flexible parameters accepted by this callable.

    Raises
    ------
    exception_cls
         If that callable is *not* pure-Python.
    '''
    assert isinstance(func, CallableFunctoolsPartialType), (
        f'{repr(func)} not "function.partial"-wrapped callable.')

    # Avoid circular import dependencies.
    from beartype._util.func.arg.utilfuncarglen import (
        get_func_args_flexible_len)

    # Pure-Python wrappee callable wrapped by that partial.
    wrappee = unwrap_func_functools_partial_once(func)

    # Positional and keyword parameters implicitly passed by this partial to
    # this wrappee.
    partial_args, partial_kwargs = get_func_functools_partial_args(func)

    # Number of flexible parameters accepted by this wrappee.
    #
    # Note that this recursive function call is guaranteed to immediately bottom
    # out and thus be safe. Why? Because a partial *CANNOT* wrap itself, because
    # a partial has yet to be defined when the functools.partial.__init__()
    # method defining that partial is called. Technically, the caller *COULD*
    # violate sanity by directly interfering with the "func" instance variable
    # of this partial after instantiation. Pragmatically, a malicious edge case
    # like that is unlikely in the extreme. You are now reading this comment
    # because this edge case just blew up in your face, aren't you!?!? *UGH!*
    wrappee_args_flexible_len = get_func_args_flexible_len(
        func=wrappee,
        is_unwrap=is_unwrap,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )

    # Number of flexible parameters passed by this partial to this wrappee.
    partial_args_flexible_len = len(partial_args) + len(partial_kwargs)

    # Number of flexible parameters accepted by this wrappee minus the number of
    # flexible parameters passed by this partial to this wrappee.
    func_args_flexible_len = (
        wrappee_args_flexible_len - partial_args_flexible_len)

    # If this number is negative, the caller maliciously defined an invalid
    # partial passing more flexible parameters than this wrappee accepts. In
    # this case, raise an exception.
    #
    # Note that the "functools.partial" factory erroneously allows callers to
    # define invalid partials passing more flexible parameters than their
    # wrappees accept. Ergo, validation is required to guarantee sanity.
    if func_args_flexible_len < 0:
        raise exception_cls(
            f'{exception_prefix}{repr(func)} passes '
            f'{partial_args_flexible_len} parameter(s) to '
            f'{repr(wrappee)} accepting only '
            f'{wrappee_args_flexible_len} parameter(s) '
            f'(i.e., {partial_args_flexible_len} > '
            f'{wrappee_args_flexible_len}).'
        )
    # Else, this number is non-negative. The caller correctly defined a valid
    # partial passing no more flexible parameters than this wrappee accepts.

    # Return this number.
    return func_args_flexible_len

# ....................{ UNWRAPPERS                         }....................
def unwrap_func_functools_partial_once(
    func: CallableFunctoolsPartialType) -> Callable:
    '''
    Possibly C-based callable directly wrapped by the passed **partial** (i.e.,
    pure-Python callable :class:`functools.partial` object directly wrapping
    this possibly C-based callable).

    Parameters
    ----------
    func : CallableFunctoolsPartialType
        Partial to be unwrapped.

    Returns
    -------
    Callable
        Possibly C-based callable directly wrapped by this partial.
    '''
    assert isinstance(func, CallableFunctoolsPartialType), (
        f'{repr(func)} not "function.partial"-wrapped callable.')

    # Return the public "func" instance variable of this partial wrapper as is.
    return func.func

# ....................{ DECORATORS                         }....................
def beartype_functools_lru_cache(
    pseudofunc: BeartypeableT, **kwargs) -> BeartypeableT:
    '''
    Monkey-patch the passed :func:`functools.lru_cache`-memoized
    **pseudo-callable** (i.e., low-level C-based callable object both created
    and returned by the standard :func:`functools.lru_cache` decorator) with
    dynamically generated type-checking.

    Parameters
    ----------
    pseudofunc : BeartypeableT
        Pseudo-callable to be monkey-patched by :func:`beartype.beartype`.

    All remaining keyword parameters are passed as is to the lower-level
    :func:`.beartype_func` decorator internally called by this higher-level
    decorator on the pure-Python function encapsulated in this descriptor.

    Returns
    -------
    BeartypeableT
        New pseudo-callable monkey-patched by :func:`beartype.beartype`.
    '''

    # Avoid circular and third-party import dependencies.
    from beartype._decor._nontype.decornontype import beartype_func
    from beartype._util.func.utilfuncwrap import unwrap_func_once

    # If this pseudo-callable is *NOT* actually a @functools.lru_cache-memoized
    # callable, raise an exception.
    if not is_func_functools_lru_cache(pseudofunc):
        raise BeartypeDecorWrappeeException(  # pragma: no cover
            f'@functools.lru_cache-memoized callable {repr(pseudofunc)} not  '
            f'decorated by @functools.lru_cache.'
        )
    # Else, this pseudo-callable is a @functools.lru_cache-memoized callable.

    # Original pure-Python callable decorated by @functools.lru_cache.
    func = unwrap_func_once(pseudofunc)  # pyright: ignore

    # Decorate that callable with type-checking.
    func_checked = beartype_func(func=func, **kwargs)

    # Dictionary mapping from the names of all keyword parameters originally
    # passed by the caller to that decorator, enabling the re-decoration of that
    # callable. Thankfully, that decorator preserves these parameters via the
    # decorator-specific "cache_parameters" instance variable whose value is a
    # bizarre argumentless lambda function (...for unknown reasons that are
    # probably indefensible) creating and returning this dictionary: e.g.,
    #     >>> from functools import lru_cache
    #     >>> @lru_cache(maxsize=3)
    #     ... def plus_one(n: int) -> int: return n +1
    #     >>> plus_one.cache_parameters()
    #     {'maxsize': 3, 'typed': False}
    lru_cache_kwargs = pseudofunc.cache_parameters()  # type: ignore[attr-defined]

    # Closure defined and returned by the @functools.lru_cache decorator when
    # passed these keyword parameters.
    lru_cache_configured = lru_cache(**lru_cache_kwargs)

    # Re-decorate that callable by @functools.lru_cache by the same parameters
    # originally passed by the caller to that decorator.
    pseudofunc_checked = lru_cache_configured(func_checked)

    # Return that new pseudo-callable.
    return pseudofunc_checked  # pyright: ignore



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype Inferential Type-hint Engine (BITE) collection items type hint
inferrers** (i.e., lower-level functions dynamically inferring subscripted type
hints describing pure-Python containers containing one or more items).
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeConfException
from beartype.typing import (
    Counter,
    Optional,
    Tuple,
)
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confenum import BeartypeStrategy
from beartype._data.typing.datatyping import FrozenSetInts
from beartype._util.hint.pep.proposal.pep484585646 import (
    make_hint_pep484585_tuple_fixed)
from beartype._util.hint.pep.proposal.pep484604 import (
    make_hint_pep484604_union)
from beartype._util.kind.integer.utilintget import (
    get_integer_pseudorandom_signed_32bit)
from collections.abc import (
    Collection as CollectionABC,
    Mapping as MappingABC,
    Sequence as SequenceABC,
)

# ....................{ INFERERS                           }....................
def infer_hint_collection_items(
    # Mandatory parameters.
    obj: CollectionABC,
    hint_factory: object,
    conf: BeartypeConf,
    __beartype_obj_ids_seen__: FrozenSetInts,

    # Optional parameters.
    origin_type: Optional[type] = None,
) -> object:
    '''
    Type hint recursively validating the passed collection (including *all*
    items transitively reachable from this collection), defined by subscripting
    the passed type hint factory by the union of the child type hints validating
    these items.

    This function *cannot* be memoized, due to necessarily accepting the
    ``__beartype_obj_ids_seen__`` parameter unique to each call to the parent
    :func:`beartype.bite.infer_hint` function.

    Parameters
    ----------
    obj : CollectionABC
        Collection to infer a type hint from.
    hint_factory : object
        Subscriptable type hint factory validating this collection (e.g., the
        :pep:`585`-compliant :class:`list` builtin type if this collection is a
        list).
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object).
    __beartype_obj_ids_seen__ : FrozenSet[int]
        **Recursion guard.** See also the parameter of the same name accepted by
        the :func:`beartype.bite.inferhint.infer_hint` function.
    origin_type : Optional[type]
        Either:

        * If the caller requires support for so-called "virtual subclasses" in
          which the passed ``obj`` collection is an instance of a user-defined
          class that does *not* explicitly subclass a :mod:`collections.abc`
          abstract base class (ABC) but does nonetheless implicitly satisfy the
          protocol implied by such an ABC, the ABC implicitly satisfied by this
          collection. Ideally, all :mod:`collections.abc` collections would
          support virtual subclassing by defining the ``__subclasshook__()`
          dunder method. Indeed, most do -- but some (e.g.,
          :class:`collections.abc.Mapping`) do *not*. This parameter enables
          callers to overcome this probably unintentional oversight in CPython.
        * Else, :data:`None`. In this case, this parameter actually defaults to
          the type of the passed ``obj`` collection.

        Defaults to :data:`None`.

    Returns
    -------
    object
        Type hint inferred from the passed collection.

    Warns
    -----
    BeartypeDoorInferHintRecursionWarning
        On detecting that the passed iterable is **recursive** (i.e.,
        containing one or more items that self-referentially refer to this same
        iterable).
    '''
    assert isinstance(obj, CollectionABC), f'{repr(obj)} not collection.'
    assert isinstance(conf, BeartypeConf), f'{repr(conf)} not configuration.'
    assert isinstance(__beartype_obj_ids_seen__, frozenset), (
        f'{repr(__beartype_obj_ids_seen__)} not frozen set.')

    # ....................{ PREAMBLE                       }....................
    # If this collection is empty, return the this unsubscripted hint factory
    # permissively matching *ALL* collections of this type. Since *NO* child
    # type hints can be safely inferred from an empty collection, our only
    # recourse is to allow similar instances of this collection to contain *ALL*
    # possible items.
    if not obj:
        return hint_factory
    # Else, this collection is non-empty.
    #
    # If no origin type was passed, default this to the type of this collection.
    elif origin_type is None:
        origin_type = obj.__class__
    # Else, an origin type was passed. Preserve this type as is.
    assert isinstance(origin_type, type), (
        f'{repr(origin_type)} not type.')

    # ....................{ LOCALS                         }....................
    # Add the integer uniquely identifying this collection to this set, thus
    # recording that this collection has now been visited by this recursion.
    __beartype_obj_ids_seen__ |= {id(obj)}

    # Low-level private callable defined below suitable for inferring the full
    # type hint recursively validating this collection, defined as either...
    hint_inferer = (
        # If this collection is a mapping, the mapping-specific inferer;
        #
        # Ideally, detecting whether a collection is a mapping would be
        # trivially feasible with a standard one-liner resembling:
        #     if isinstance(obj, collections.abc.Mapping) else
        #
        # Unfortunately, the "collections.abc.Mapping" ABC is *BROKEN.* Unlike
        # most other "collections.abc" superclasses, "Mapping" fails to define
        # the __subclasshook__() dunder method and thus fails to support the
        # so-called "virtual subclasses" that most "collections.abc"
        # superclasses support. See also this relevant StackOverflow answer:
        #     https://stackoverflow.com/a/64666157/2809027
        #
        # Our only recourse is to require that callers requiring support for
        # "virtual subclasses" pass a distinct "hint_origin_type" class that
        # can then be tested as a "collections.abc.Mapping" subclass.
        _infer_hint_mapping_items
        if issubclass(origin_type, MappingABC) else
        # Else, this collection is *NOT* a mapping. In this case, the
        # general-purpose inferer applicable to *ALL* single-argument
        # reiterables (e.g., collections whose type hint factories are
        # subscriptable by only a single child type hint).
        _infer_hint_reiterable_items
    )
    # print(f'Inferring {repr(obj)} child hints with {repr(hint_inferer)}...')

    # Type hint recursively validating this collection.
    hint = hint_inferer(
        obj=obj,  # type: ignore[arg-type]
        hint_factory=hint_factory,
        conf=conf,
        __beartype_obj_ids_seen__=__beartype_obj_ids_seen__,
    )

    # Return this hint.
    return hint

# ....................{ PRIVATE ~ constants                }....................
_ROOT_TUPLE_FIXED_ITEMS_LEN_MAX = 10
'''
Maximum inclusive number of tuple items below which the private
:func:`._infer_hint_reiterable_items` function infers a **root tuple** (i.e.,
top-most object originally passed by the caller to the public
:func:`beartype.bite.infer_hint` function) to be validated by a **fixed-length
tuple type hint** of the form ``tuple[{hint_child1}, ???, {hint_childN}]``.

Specifically, for each root tuple containing:

* Less than or equal to this number of items, this tuple is annotated by a
  **fixed-length tuple type hint** of the form ``tuple[{hint_child1}, ???,
  {hint_childN}]``.
* Greater than this number of items, this tuple is annotated by a **variadic
  tuple type hint** of the form ``tuple[{hint_childs}, ...]``.

This magic number enables an ad-hoc heuristic for disambiguating between
fixed-length and variadic tuple type hints. Technically, *any* tuple may be
ambiguously annotated as either. Pragmatically, fixed-length tuple type hints
exist almost exclusively to annotate **multiple-return functions** (i.e.,
functions returning two or more values as a tuple whose items are those values);
variadic tuple type hints annotate all other tuples, which is most of them.
Since all multiple-return functions return a root tuple *and* since most
real-world multiple-return functions of interest return a root tuple containing
less than or equal to this magic number of items, this heuristic follows.
'''

# ....................{ PRIVATE ~ inferers                 }....................
def _infer_hint_mapping_items(
    obj: MappingABC,
    hint_factory: object,
    conf: BeartypeConf,
    __beartype_obj_ids_seen__: FrozenSetInts,
) -> object:
    '''
    Type hint recursively validating the passed **mapping** (i.e.,
    collections whose type hint factories are subscriptable by a pair of child
    key and value type hints) and all keys *and* values transitively reachable
    from this mapping, defined by subscripting the passed type hint factory by
    the unions of the child type hints validating these keys and values.

    See Also
    --------
    :func:`.infer_hint_collection_items`
        Further details.
    '''
    # Note that we intentionally do *NOT* test whether this object also
    # satisfies the more fine-grained "collections.abc.Mapping" protocol. Why?
    # Because that protocol fails to support so-called "virtual subclasses" as
    # of Python 3.13 and is thus fundamentally broken. See also the parent
    # infer_hint_collection_items() function "origin_type" parameter. *sigh*

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype.bite._infermain import infer_hint

    # ....................{ LOCALS                         }....................
    # Child key and value hints to subscript this factory with, defaulting to
    # the ignorable "object" superclass.
    hints_key: object = object
    hints_value: object = object

    # ....................{ RECURSION                      }....................
    # Note that the caller guarantees this mapping to be non-empty.

    # If either...
    if (
        # This mapping contains only one key-value pair *OR*...
        #
        # Note that this is a negligible optimization for this common case.
        len(obj) == 1 or
        # The caller requested O(1) constant-time complexity...
        conf.strategy is BeartypeStrategy.O1
    ):
        # First key-value pair of this mapping.
        key, value = next(iter(obj.items()))

        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # CAUTION: Synchronize with similar logic below.
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # Child key and value type hints validating this key and value.
        hints_key = infer_hint(
            obj=key,
            conf=conf,
            __beartype_obj_ids_seen__=__beartype_obj_ids_seen__,
        )
        hints_value = infer_hint(
            obj=value,
            conf=conf,
            __beartype_obj_ids_seen__=__beartype_obj_ids_seen__,
        )
    # Else, the caller did *NOT* request O(1) constant-time complexity.
    #
    # If the caller requested O(n) linear-time complexity...
    elif conf.strategy is BeartypeStrategy.On:
        # Set of all child key and value hints to conjoin into a union.
        hints_key_set = set()
        hints_value_set = set()

        # For each key-value pair in this mapping...
        for key, value in obj.items():
            #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            # CAUTION: Synchronize with similar logic above.
            #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            # Child key and value type hints validating this key and value.
            hint_key = infer_hint(
                obj=key,
                conf=conf,
                __beartype_obj_ids_seen__=__beartype_obj_ids_seen__,
            )
            hint_value = infer_hint(
                obj=value,
                conf=conf,
                __beartype_obj_ids_seen__=__beartype_obj_ids_seen__,
            )

            # Add this child type hint to this set.
            hints_key_set.add(hint_key)
            hints_value_set.add(hint_value)

        # PEP 604- or 484-compliant union of these child key and value hints.
        hints_key = make_hint_pep484604_union(tuple(hints_key_set))
        hints_value = make_hint_pep484604_union(tuple(hints_value_set))
    # Else, the caller did *NOT* request O(n) linear-time complexity.
    #
    # Thus, the caller requested a currently unsupported time complexity. In
    # this case, raise an exception.
    else:
        raise BeartypeConfException(
            f'Beartype configuration {repr(conf)} '
            f'strategy {repr(conf.strategy)} currently unsupported by '
            f'beartype.bite.infer_hint() (i.e., neither '
            f'"BeartypeStrategy.O1" nor "BeartypeStrategy.On").'
        )

    # ....................{ SUBSCRIPTION                   }....................
    # If these child key and value hints are still the ignorable "object"
    # superclass, return this factory as is unsubscripted.
    if hints_key is object and hints_value is object:
        hint = hint_factory
    # Else, at least one of these child key or value hints is no longer the
    # ignorable "object" superclass. In this case...
    else:
        # Type hint recursively validating this mapping, defined as either...
        hint = (
            # If this mapping is a counter (i.e., instance of the standard
            # "collections.Counter" class), subscripting this factory by only
            # this key union. By definition, *ALL* values of *ALL* counters are
            # unconditionally constrained to be integers and thus need *NOT*
            # (and indeed *CANNOT*) be explicitly specified;
            hint_factory[hints_key]  # type: ignore[index]
            if hint_factory is Counter else
            # Else, this mapping is *NOT* a counter. In this case, sequentially
            # subscripting this factory by both this key and value union.
            hint_factory[hints_key, hints_value]  # type: ignore[index]
        )
    # print(f'Inferred {repr(obj)} hint as {repr(hint)}...')

    # Return this hint.
    return hint


def _infer_hint_reiterable_items(
    obj: CollectionABC,
    hint_factory: object,
    conf: BeartypeConf,
    __beartype_obj_ids_seen__: FrozenSetInts,
) -> object:
    '''
    Type hint recursively validating the passed **reiterable** (i.e.,
    collections whose type hint factories are subscriptable by only a single
    child type hint) and all items transitively reachable from this reiterable,
    defined by subscripting the passed type hint factory by the union of the
    child type hints validating these items.

    See Also
    --------
    :func:`.infer_hint_collection_items`
        Further details.
    '''

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype.bite._infermain import infer_hint

    # ....................{ RECURSION                      }....................
    # Note that the caller guarantees this mapping to be non-empty.

    # If...
    if (
        # This collection is a tuple *AND*...
        hint_factory is Tuple and
        # The frozen set of the object IDs of all currently inferred parent
        # collections contains exactly one ID, this is a root tuple (i.e.,
        # top-most tuple originally passed by the caller to the parent
        # infer_hint() function). If this is the case *AND*...
        len(__beartype_obj_ids_seen__) == 1 and
        # This root tuple contains less than or equal to this ad-hoc maximum
        # number of tuple items...
        len(obj) <= _ROOT_TUPLE_FIXED_ITEMS_LEN_MAX
    ):
        # List of all child type hints validating the items of this tuple.
        hints_item_list = []

        # For each item in this tuple...
        #
        # Note that we intentionally ignore the "conf.strategy" option
        # providing caller's preferred time complexity in this edge case. Why?
        # Because:
        # * This behaviour mirrors that of the *ALL* other beartype runtime
        #   type-checkers, which similarly ignore the "conf.strategy" option
        #   when checking fixed-length tuple type hints.
        # * This tuple is guaranteed to be small, effectively reducing this
        #   iteration to O(1) constant-time complexity in any case.
        for item in obj:
            # Child type hint validating this item.
            hint_item = infer_hint(
                obj=item,
                conf=conf,
                __beartype_obj_ids_seen__=__beartype_obj_ids_seen__,
            )

            # Append this child type hint to this list.
            hints_item_list.append(hint_item)

        # Then this root tuple is subjectively small enough to be returned as
        # the value returned by a multiple-return function (i.e., returning two
        # or more values as items of this tuple). In this case, validate this
        # tuple with a fixed-length tuple type hint. See the
        # "_ROOT_TUPLE_FIXED_ITEMS_LEN_MAX" docstring for further details.
        hint = make_hint_pep484585_tuple_fixed(tuple(hints_item_list))
    # Else, this collection is *NOT* a subjectively small root tuple. In this
    # case, defer to generic logic.
    else:
        # Child item hint to subscript this factory with, defaulting to the
        # ignorable "object" superclass.
        hints_item: object = object

        # If either...
        if (
            # This reiterable contains only one item *OR*...
            #
            # Note that this is a negligible optimization for this common case.
            len(obj) == 1 or
            # The caller requested O(1) constant-time complexity...
            conf.strategy is BeartypeStrategy.O1
        ):
            # Item of this reiterable to infer the child type hint of.
            item: object = None  # type: ignore[no-redef]

            # If this reiterable is a sequence, this reiterable supports
            # efficient random access to arbitrary items. In this case...
            if isinstance(obj, SequenceABC):
                # Pseudorandom signed 32-bit integer.
                int_random = get_integer_pseudorandom_signed_32bit()

                # 0-based pseudorandom index into this sequence.
                obj_item_index_random = int_random % len(obj)

                # Pseudorandom item of this sequence.
                item = obj[obj_item_index_random]
            # Else, this reiterable is *NOT* a sequence, implying this
            # reiterable fails to support efficient random access to arbitrary
            # items. In this case...
            else:
                # First item of this reiterable.
                item = next(iter(obj))

            # Child type hint validating this item.
            hints_item = infer_hint(
                obj=item,
                conf=conf,
                __beartype_obj_ids_seen__=__beartype_obj_ids_seen__,
            )
        # Else, the caller did *NOT* request O(1) constant-time complexity.
        #
        # If the caller requested O(n) linear-time complexity...
        elif conf.strategy is BeartypeStrategy.On:
            # List of all child type hints validating the items of this reiterable.
            hints_item_list = []

            # For each item in this reiterable...
            for item in obj:
                # Child type hint validating this item.
                hint_item = infer_hint(
                    obj=item,
                    conf=conf,
                    __beartype_obj_ids_seen__=__beartype_obj_ids_seen__,
                )

                # Append this child type hint to this list.
                hints_item_list.append(hint_item)

            # Attempt to temporarily coerce this list of child type hints into a
            # set to trivially remove *ALL* duplicate hints.
            try:
                hints_item_list = set(hints_item_list)  # type: ignore[assignment]
            # If doing so raises *ANY* exception whatsoever, silently ignore this
            # exception. In all likelihood, this exception connotes a hashability
            # error due to one or more of these child type hints being unhashable
            # and thus *NOT* addable to a set. Although non-ideal, this certainly
            # isn't worth destroying type hint inference over.
            except Exception:
                pass

            # PEP 604- or 484-compliant union of these child type hints.
            hints_item = make_hint_pep484604_union(tuple(hints_item_list))
        # Else, the caller did *NOT* request O(n) linear-time complexity.
        #
        # Thus, the caller requested a currently unsupported time complexity. In
        # this case, raise an exception.
        else:
            raise BeartypeConfException(
                f'Beartype configuration {repr(conf)} '
                f'strategy {repr(conf.strategy)} currently unsupported by '
                f'beartype.bite.infer_hint() (i.e., neither '
                f'"BeartypeStrategy.O1" nor "BeartypeStrategy.On").'
            )

        # ....................{ SUBSCRIPTION               }....................
        # If this child hint is still the ignorable "object" superclass, return
        # this factory as is unsubscripted.
        if hints_item is object:
            hint = hint_factory
        # Else, this child hint is no longer the ignorable "object" superclass.
        # In this case...
        else:
            # Type hint recursively validating this reiterable, defined by...
            hint = (
                # If this collection is a tuple, subscripting this tuple hint
                # factory by the variadic-length variant of this union followed
                # by an ellipses (signifying this tuple to contain arbitrarily
                # many items);
                hint_factory[hints_item, ...]  # type: ignore[index]
                if hint_factory is Tuple else
                # Else, this collection is *NOT* a tuple. In this case,
                # subscripting this non-tuple hint factory by this union.
                hint_factory[hints_item]  # type: ignore[index]
            )

    # Return this hint.
    return hint



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`563` **unit tests.**

This submodule unit tests the public :func:`beartype.peps.resolve_pep563`
function, which internally leverages the :mod:`beartype.door` subpackage to
validate itself and is thus intentionally deferred to this DOOR-specific test
subpackage.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# .....................{ TESTS                             }....................
def test_resolve_pep563() -> None:
    '''
    Test the :func:`beartype.peps.resolve_pep563` resolver.
    '''

    # .....................{ IMPORTS                       }....................
    # Defer test-specific imports.
    from beartype.peps import resolve_pep563
    from beartype.roar import (
        BeartypeCallHintForwardRefException,
        BeartypeDecorHintForwardRefException,
        BeartypePep563Exception,
    )
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_MOST_3_13
    from beartype_test.a00_unit.data.pep.pep563.data_pep563_resolve import (
        ToAvariceOrPride,
        FrequentWith,
        their_starry_domes,
    )
    from pytest import raises

    # .....................{ LOCALS                        }....................
    # Arbitrary instance of the generic accepted by the function called below.
    numberless_and_immeasurable_halls = ToAvariceOrPride()

    # Arbitrary instance of the class defining various problematic methods
    # called below.
    and_thrones_radiant_with_chrysolite = FrequentWith()

    # .....................{ FAIL                          }....................
    # Assert that this function unsuccessfully raises the expected exception
    # *BEFORE* resolving all PEP 563-postponed type hints annotating these
    # callables.
    with raises(BeartypeDecorHintForwardRefException):
        their_starry_domes(numberless_and_immeasurable_halls)
    with raises(BeartypeDecorHintForwardRefException):
        and_thrones_radiant_with_chrysolite.until_the_doves(
            numberless_and_immeasurable_halls)
    with raises(BeartypeDecorHintForwardRefException):
        and_thrones_radiant_with_chrysolite.crystal_column(
            'Nor had that scene of ampler majesty')

    # .....................{ PASS                          }....................
    # Resolve all PEP 563-postponed type hints annotating these callables.
    resolve_pep563(their_starry_domes)
    resolve_pep563(FrequentWith.until_the_doves)
    resolve_pep563(FrequentWith.crystal_column)

    # Assert that this function successfully accepts and returns this instance.
    assert their_starry_domes(numberless_and_immeasurable_halls) is (
        numberless_and_immeasurable_halls)

    # Assert that this method successfully accepts and returns this string.
    assert FrequentWith.until_the_doves(numberless_and_immeasurable_halls) is (
        numberless_and_immeasurable_halls)

    # .....................{ FAIL ~ more                   }....................
    # Assert that this resolver raises the expected exception when passed an
    # uncallable object.
    with raises(BeartypePep563Exception):
        resolve_pep563('Mont Blanc yet gleams on high:—the power is there,')

    # Assert that this method unsuccessfully raises the excepted exception, due
    # to being annotated by a missing forward reference.
    with raises(BeartypeCallHintForwardRefException):
        and_thrones_radiant_with_chrysolite.crystal_column(
            'Than gems or gold, the varying roof of heaven')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`484`-compliant **type variable type-checking code factories**
(i.e., low-level callables dynamically generating pure-Python code snippets
type-checking arbitrary objects against type variables).

This private submodule is *not* intended for importation by downstream callers.
'''

#FIXME: Excise us up but preserve this submodule, please. We'll want this once
#we begin generating call-time code type-checking type variables.

# # ....................{ IMPORTS                            }....................
# from beartype.typing import Optional
# from beartype._check.code.codecls import HintMeta
# from beartype._check.convert.convmain import (
#     sanify_hint_child)
# from beartype._conf.confmain import BeartypeConf
# from beartype._data.typing.datatypingport import Hint
# from beartype._data.typing.datatyping import TypeStack
# from beartype._util.hint.pep.proposal.pep484.pep484typevar import (
#     get_hint_pep484_typevar_bounded_constraints_or_none)
#
# # ....................{ FACTORIES                          }....................
# def make_hint_pep484_typevar_check_expr(
#     # Mandatory parameters.
#     hint_meta: HintMeta,
#     conf: BeartypeConf,
#     pith_curr_assign_expr: str,
#     pith_curr_var_name_index: int,
#
#     # Optional parameters.
#     cls_stack: TypeStack = None,
#     exception_prefix: str = '',
# ) -> Optional[Hint]:
#     '''
#     Either reduce the mostly semantically useless passed :pep:`484`-compliant
#     **type variable** (i.e., :class:`typing.TypeVar` object) to the semantically
#     useful bounds or constraints of this variable if any *or* silently ignore
#     this variable otherwise (i.e., if this variable is neither bounded nor
#     constrained)..
#
#     This factory is intentionally *not* memoized (e.g., by the
#     :func:`.callable_cached` decorator), due to accepting **context-sensitive
#     parameters** (i.e., whose values contextually depend on context unique to
#     the code being generated for the currently decorated callable) such as
#     ``pith_curr_assign_expr``.
#
#     Parameters
#     ----------
#     hint_meta : HintMeta
#         Metadata describing the currently visited hint, appended by the
#         previously visited parent hint to the ``hints_meta`` stack.
#     conf : BeartypeConf
#         **Beartype configuration** (i.e., self-caching dataclass encapsulating
#         all settings configuring type-checking for the passed object).
#     pith_curr_assign_expr : str
#         Assignment expression assigning this full Python expression to the
#         unique local variable assigned the value of this expression.
#     pith_curr_var_name_index : int
#         Integer suffixing the name of each local variable assigned the value of
#         the current pith in a assignment expression, thus uniquifying this
#         variable in the body of the current wrapper function.
#     cls_stack : TypeStack, optional
#         **Type stack** (i.e., either a tuple of the one or more
#         :func:`beartype.beartype`-decorated classes lexically containing the
#         class variable or method annotated by this hint *or* :data:`None`).
#         Defaults to :data:`None`.
#     exception_prefix : str, optional
#         Human-readable substring prefixing the representation of this object in
#         the exception message. Defaults to the empty string.
#
#     Returns
#     -------
#     Optional[Hint]
#         Either:
#
#         * If this type variable reduces to an unignorable type hint, that hint.
#         * Else, :data:`None` (i.e., if this type variable is ignorable).
#     '''
#     assert isinstance(hint_meta, HintMeta), (
#         f'{repr(hint_meta)} not "HintMeta" object.')
#
#     # ....................{ LOCALS                         }....................
#     # This type variable, localized for negligible efficiency gains. *sigh*
#     hint = hint_meta.hint
#
#     # Hint mapped to by this type variable if one or more transitive parent
#     # hints previously mapped this type variable to a hint *OR* "None".
#     hint_child: Optional[Hint] = hint_meta.typearg_to_hint.get(hint)  # pyright: ignore
#
#     # ....................{ REDUCTION                      }....................
#     # If *NO* transitive parent hints previously mapped this type variable to a
#     # hint...
#     if hint_child is None:
#         # PEP-compliant hint synthesized from all bounded constraints
#         # parametrizing this type variable if any *OR* "None" otherwise.
#         #
#         # Note this call is intentionally passed positional rather positional
#         # keywords due to memoization.
#         hint_curr_bound = get_hint_pep484_typevar_bounded_constraints_or_none(
#             hint, exception_prefix)  # pyright: ignore
#
#         # If this type variable was parametrized by one or more bounded
#         # constraints, reduce this type variable to these bounded constraints.
#         if hint_curr_bound is not None:
#             hint_child = hint_curr_bound
#         # Else, this type variable was unparametrized. In this case, preserve
#         # this type variable as is.
#     # Else, one or more transitive parent hints previously mapped this type
#     # variable to a hint.
#
#     # If this type variable was either mapped to another hint by one or more
#     # transitive parent hints *OR* parametrized by one or more bounded
#     # constraints...
#     if hint_child is not None:
#         # Unignorable sane hint sanified from this possibly ignorable insane
#         # hint *OR* "None" otherwise (i.e., if this hint is ignorable).
#         hint_child = sanify_hint_child(
#             hint=hint_child,
#             conf=conf,
#             cls_stack=cls_stack,
#             exception_prefix=exception_prefix,
#         )
#     # Else, this type variable was neither mapped to another hint by one or more
#     # transitive parent hints *NOR* parametrized by one or more bounded
#     # constraints.
#
#     # If this type variable is reducible to an unignorable hint...
#     if hint_child is not None:
#         # Ignore this semantically useless type variable in favour of this
#         # semantically useful hint by replacing *ALL* hint metadata describing
#         # the former with the latter.
#         hint_meta.reinit(
#             hint=hint_child,  # pyright: ignore
#             indent_level=hint_meta.indent_level + 1,
#             pith_expr=pith_curr_assign_expr,
#             pith_var_name_index=pith_curr_var_name_index,
#             typearg_to_hint=hint_meta.typearg_to_hint,
#         )
#     # Else, this type variable is *NOT* reducible to an unignorable hint. Since
#     # @beartype currently fails to generate type-checking code for type
#     # variables in and of themselves, type variables have *NO* intrinsic
#     # semantic meaning and are thus ignorable.
#
#     # ....................{ RETURN                         }....................
#     # Return either the unignorable hint this type variable reduces to if any
#     # *OR* "None" otherwise (i.e., if this type variable is ignorable).
#     return hint_child



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **intraprocess import hook FastMCP integration tests** (i.e.,
exercising :mod:`beartype.claw` import hooks specific to the third-party
FastMCP API within the active Python process).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: Isolate each integration test defined below to its own subprocess.
# Why? Module imports. Since each integration test defined below tends to
# reimport the same (or, at least, similar) modules as previously run unit tests
# defined below, module imports and thus unit tests *MUST* be isolated to their
# own subprocesses to ensure these tests may be run in any arbitrary order.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
import pytest
from beartype_test._util.mark.pytskip import (
    skip_unless_os_linux,
    skip_unless_package,
)

# ....................{ TESTS                              }....................
# Explicitly skip this integration test under both vanilla Microsoft Windows
# *AND* macOS, where this test inexplicably erupts in unreadable pickling errors
# resembling:
#     _pickle.PicklingError: Can't pickle <function test_claw_fastmcp at
#     0x000001DCDB47AA20>: it's not the same object as
#     beartype_test.a90_func.a40_claw.external.test_claw_fastmcp.test_claw_fastmcp
#
# Clearly, this relates to coroutines. Private OS-specific implementations of
# the standard "pickle" module fail to cope with coroutines, apparently. *shrug*
@skip_unless_package('fastmcp')
@skip_unless_os_linux()
@pytest.mark.run_in_subprocess
async def test_claw_fastmcp() -> None:
    '''
    Test :mod:`beartype.claw` import hooks against a FastMCP-specific data
    subpackage in this test suite exercising these hooks against the third-party
    FastMCP API known to publish decorator-hostile decorators.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.claw import beartype_package

    # ....................{ PASS                           }....................
    # Subject a FastMCP-specific data package to a beartype import hook,
    # configured by the default beartype configuration.
    beartype_package('beartype_test.a90_func.data.claw.external.fastmcp')

    # Import all submodules of the package hooked above, exercising that these
    # submodules are subject to that import hook.
    from beartype_test.a90_func.data.claw.external.fastmcp.data_claw_fastmcp import (
        data_claw_fastmcp_main,
        with_stride_colossal,
    )

    # Asynchronously run the main public coroutine exported by that submodule,
    # thus asserting that *ALL* integration tests (implemented as assertion
    # statements) in that submodule are satisfied.
    await data_claw_fastmcp_main()
    # print('[test_claw_fastmcp] with_stride_colossal:')
    # print(with_stride_colossal.__module__)
    # print(with_stride_colossal.__class__.__name__)

    # Implicitly assert that the @beartype decorator avoids raising exceptions
    # (typically by internally reducing to a noop) when decorating uncallable
    # objects produced by FastMCP-specific decorator-hostile decorators.
    with_stride_colossal_beartyped = beartype(with_stride_colossal)

    # Assert that the @beartype decorator actually did reduce to a noop.
    assert with_stride_colossal is with_stride_colossal_beartyped



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`-compliant **type variable utilities** (i.e.,
callables generically applicable to :pep:`484`-compliant type variables).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep484TypeVarException
from beartype.typing import TypeVar
from beartype._data.kind.datakindiota import SENTINEL
from beartype._data.typing.datatyping import TypeException
from beartype._data.typing.datatypingport import (
    Hint,
    HintOrNone,
    TypeIs,
)
from beartype._util.cache.utilcachecall import callable_cached

# ....................{ RAISERS                            }....................
#FIXME: Unit test us up, please. *sigh*
def die_unless_hint_pep484_typevar(
    # Mandatory parameters.
    hint: TypeVar,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep484TypeVarException,
    exception_prefix: str = '',
) -> None:
    '''
    Raise the passed exception unless the passed type hint is a
    :pep:`484`-compliant **type variable** (i.e., :class:`.TypeVar` object).

    Parameters
    ----------
    hint : TypeVar
        Type hint to be inspected.
    exception_cls : TypeException, default: BeartypeDecorHintPep484TypeVarException
        Type of exception to be raised. Defaults to
        :exc:`.BeartypeDecorHintPep484TypeVarException`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Raises
    ------
    exception_cls
        If this hint is *not* a type variable.
    '''

    # If this hint is *NOT* a type variable...
    if not is_hint_pep484_typevar(hint):  # pyright: ignore
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not type.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        # Raise an exception of this type prefixed by this prefix.
        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} '
            f'not PEP 484 type variable.'
        )
    # Else, this hint is a type variable.

# ....................{ TESTERS                            }....................
#FIXME: Unit test us up, please.
def is_hint_pep484_typevar(hint: Hint) -> TypeIs[TypeVar]:  # pyright: ignore
    '''
    :data:`True` only if the passed object is a :pep:`484`-compliant **type
    variable** (i.e., :class:`typing.TypeVar` object).
    '''

    # Although this test currently reduces to a trivial one-liner, it's *NOT*
    # inconceivable that this test could become non-trivial under a subsequent
    # CPython version. *shrug*
    return isinstance(hint, TypeVar)

# ....................{ GETTERS                            }....................
#FIXME: *UHM.* We probably shouldn't be treating type variable constraints and
#upper bounds as semantically equivalent. They're really not at all. We suspect
#our younger self was simply confused by the frankly unreadable PEP
#documentation surrounding this subject. In a readable nutshell:
#* Type variable *BOUNDS* are arbitrary type hints. An object satisfies a type
#  variable bound if and only if that object is an instance of that bound (i.e.,
#  the type hint of that object is a subhint of that bound). Subclasses are thus
#  permitted. Standard code generation thus suffices here.
#* Type variable *CONSTRAINTS* are a tuple of one or more... what exactly?
#  Types? Type hints? If merely types, constraints are trivial to support. If
#  arbitrary type hints, however, constraints are definitely non-trivial to
#  support. We suspect that, as with bounds, constraints may be arbitrary type
#  hints. *sigh*
#
#  In either case, the distinction is clear: an object satisfies a type variable
#  constraint if and only if that object is *EXACTLY* matched by one of those
#  constraints (i.e., the type hint of that object is exactly one of those
#  constraints). Subclasses are thus prohibited. This is non-trivial to support,
#  as the only means of performing exact type hint matching is probably to:
#  * Call "beartype.bite.infer_hint(obj)" to infer the type hint for an object.
#  * Iteratively test whether that hint is one of these constraints.
#
#  That's pretty awful, though. We really shouldn't be calling infer_hint() on
#  arbitrary objects at type-checking time.
#
#  An alternative presents itself. It's extremely non-trivial, but something we
#  probably need to do anyway to support contravariant, covariant, and invariant
#  type variables. If you consider the formal definitions of bounds and
#  constraints, it's clear that:
#  * Bounds are type-checked *COVARIANTLY.* This is what almost everyone wants.
#    It's also how isinstance() and issubclass() work. So, it's how runtime
#    type-checkers behave out-of-the-box. Easy.
#  * Constraints are type-checked *INVARIANTLY.*
#
#  First, consider how our code generation algorithm currently type-checks
#  types: it assumes covariance everywhere by calling isinstance() and
#  issubclass() to perform type-checks. But what if we instead augmented our
#  make_check_expr() factory to internally track what the current kind of
#  type-checking variance is? Notably, we could:
#  * Define a new "CheckVariance" enumeration somewhere resembling:
#        class CheckVariance(Enum):
#            COVARIANCE = auto()  # <-- standard default, thus first
#            CONTRAVARIANCE = auto()
#            INVARIANCE = auto()
#  * Add a new "check_variance: CheckVariance = CheckVariance.COVARIANCE"
#    instance variable to our existing "HintMeta" dataclass.
#  * Refactor this get_hint_pep484_typevar_bounded_constraints_or_none() to return a 2-tuple
#    "(typevar_bound, check_variance)" where:
#    * "typevar_bound" is the current return value.
#    * "check_variance" is either:
#      * "None" if this type variable is unbounded and unconstrained.
#      * "CheckVariance.COVARIANCE" if this type variable is bounded.
#      * "CheckVariance.INVARIANCE" if this type variable is constrained.
#  * Refactor make_check_expr() to:
#    * Track this new "HintMeta.check_variance" instance variable. Mostly
#      trivial. Child hints inherit their parent hint's "check_variance".
#    * Dynamically generate code appropriate for this new
#      "HintMeta.check_variance" instance variable. That's the super-hard part.
#      Even if we initially ignore "CheckVariance.CONTRAVARIANCE" (which we
#      really should), we'd still basically have to:
#      * Refactor *MOST* single magic code string globals into dictionary
#        globals mapping from all possible "CheckVariance" members to the
#        corresponding magic code strings appropriate for those variances.
#      * Lookup the appropriate magic code strings in those dictionaries with
#        the current "hint_curr_meta.check_variable" instance variable.
#
#  Kinda fun, but *REALLY* non-trivial -- and probably no one cares. Guh!
@callable_cached
def get_hint_pep484_typevar_bounded_constraints_or_none(
    # Mandatory parameters.
    hint: TypeVar,

    # Optional parameters.
    exception_prefix: str = '',
) -> HintOrNone:
    '''
    PEP-compliant type hint synthesized from all bounded constraints
    parametrizing the passed :pep:`484`-compliant **type variable** (i.e.,
    :class:`typing.TypeVar` instance) if any *or* :data:`None` otherwise (i.e.,
    if this type variable was parametrized by *no* bounded constraints).

    Specifically, if this type variable was parametrized by:

    #. One or more **constraints** (i.e., positional arguments passed by the
       caller to the :meth:`typing.TypeVar.__init__` call initializing this
       type variable), this getter returns a new **PEP-compliant union type
       hint** (e.g., :attr:`typing.Union` subscription) of those constraints.
    #. One **upper bound** (i.e., ``bound`` keyword argument passed by the
       caller to the :meth:`typing.TypeVar.__init__` call initializing this
       type variable), this getter returns that bound as is.
    #. Else, this getter returns the :data:`None` singleton.

    Caveats
    -------
    **This getter treats constraints and upper bounds as semantically
    equivalent,** preventing callers from distinguishing between these two
    technically distinct variants of type variable metadata.

    For runtime type-checking purposes, type variable constraints and bounds are
    sufficiently similar as to be semantically equivalent for all intents and
    purposes. To simplify handling of type variables, this getter ambiguously
    aggregates both into the same tuple.

    For static type-checking purposes, type variable constraints and bounds are
    *still* sufficiently similar as to be semantically equivalent for all
    intents and purposes. Any theoretical distinction between the two is likely
    to be lost on *most* engineers, who tend to treat the two interchangeably.
    To quote :pep:`484`:

        ...type constraints cause the inferred type to be _exactly_ one of the
        constraint types, while an upper bound just requires that the actual
        type is a subtype of the boundary type.

    Inferred types are largely only applicable to static type-checkers, which
    internally assign type variables contextual types inferred from set and
    graph theoretic operations on the network of all objects (nodes) and
    callables (edges) relating those objects. Runtime type-checkers have *no*
    analogous operations, due to runtime space and time constraints.

    This getter is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator). If this type variable was parametrized by
    one or more constraints, the :attr:`typing.Union` type hint factory already
    caches these constraints; else, this getter performs no work. In any case,
    this getter effectively performs to work.

    Parameters
    ----------
    hint : object
        :pep:`484`-compliant type variable to be inspected.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message. Defaults to the empty string.

    Returns
    -------
    Optional[Hint]
        Either:

        * If this type variable was parametrized by one or more constraints, a
          new PEP-compliant union type hint aggregating those constraints.
        * If this type variable was parametrized by an upper bound, that bound.
        * Else, :data:`None`.

    Raises
    ------
    BeartypeDecorHintPep484TypeVarException
        if this object is *not* a :pep:`484`-compliant type variable.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.proposal.pep484604 import (
        make_hint_pep484604_union)
    from beartype._util.hint.pep.proposal.pep749 import (
        get_hint_pep749_subhint_mandatory,
        get_hint_pep749_subhint_optional,
    )

    # If this hint is *NOT* a type variable, raise an exception.
    die_unless_hint_pep484_typevar(
        hint=hint, exception_prefix=exception_prefix)
    # Else, this hint is a type variable.

    # Bounded constraints parametrizing this type variable to be returned.
    hint_bounded_constraints: Hint = None

    # Tuple of the zero or more child hints constraining this type variable.
    hint_constraints = get_hint_pep749_subhint_mandatory(
        hint=hint,  # pyright: ignore
        subhint_name_dynamic='evaluate_constraints',
        subhint_name_static='__constraints__',
        exception_cls=BeartypeDecorHintPep484TypeVarException,
        exception_prefix=exception_prefix,
    )

    #
    # If this type variable was parametrized by one or more constraints, create
    # and return the PEP 484-compliant union of these constraints.
    if hint_constraints:
        hint_bounded_constraints = make_hint_pep484604_union(hint_constraints)  # pyright: ignore
    # Else, this type variable was parametrized by *NO* constraints. In this
    # case...
    else:
        # Child hint binding this type parameter if this type parameter has a
        # bound *OR* the sentinel placeholder otherwise.
        #
        # Note that:
        # * Constraints are effectively mandatory (i.e., *ALWAYS* specified),
        #   due to the low-level C implementation of the "TypeVar" type
        #   defaulting unspecified constraints to the empty tuple -- a
        #   technically valid constraint matching zero child hints.
        # * Bounds are effectively optional (i.e., possibly unspecified), due to
        #   the low-level C implementation of the "TypeVar" type defaulting
        #   unspecified bounds to the "None" singleton. Since "None" is *NOT* a
        #   type, there exist no subclasses and thus subhints that match "None".
        #   Ergo, a type variable bound to "None" would prevent that type
        #   variable from matching any hints other than "None". That type
        #   variable would thus be equivalent to "None", which is useless,
        #   because "None" could (and should) just be used instead. PEP 484
        #   thus treats "None" as an invalid bound synonymous with *NO* bound.
        hint_bound = get_hint_pep749_subhint_optional(
            hint=hint,  # pyright: ignore
            subhint_name_dynamic='evaluate_bound',
            subhint_name_static='__bound__',
            subhint_value_null=None,
            exception_cls=BeartypeDecorHintPep484TypeVarException,
            exception_prefix=exception_prefix,
        )

        # If this type variable was parametrized by a bound, return this bound.
        if hint_bound is not SENTINEL:
            hint_bounded_constraints = hint_bound  # pyright: ignore
        # Else, this type variable was parametrized by *NO* bound.
    # Else, this type variable was parametrized by *NO* constraints.

    # Return these bounded constraints.
    return hint_bounded_constraints



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) generic type hint
classes** (i.e., :class:`beartype.door.TypeHint` subclasses implementing support
for :pep:`484`- and :pep:`585`-compliant classes subclassing subclassable type
hints, including the ``typing.Generic[...]`` and ``typing.Protocol[...]`` type
hint superclasses).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.door._cls.doorsuper import TypeHint
from beartype.door._cls.pep.pep484.doorpep484class import ClassTypeHint
from beartype.door._cls.pep.pep484585.doorpep484585subscripted import (
    SubscriptedTypeHint as SubscriptedTypeHint)
from beartype.roar import BeartypeDoorIsSubhintException
from beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget import (
    get_hint_pep484585_generic_args_full)

# ....................{ SUBCLASSES                         }....................
class GenericTypeHint(TypeHint):
    '''
    **Generic type hint wrapper** (i.e., high-level object encapsulating a
    low-level subclass of the :pep:`484`-compliant :class:`typing.Generic`
    superclass, :pep:`544`-compliant :class:`typing.Protocol` superclass, or any
    :pep:`585`-compliant subscripted type hint (e.g., ``class
    GenericListOfStrs(list[str]): ...``).
    '''

    # ..................{ PRIVATE ~ testers                  }..................
    def _is_subhint_branch(self, branch: TypeHint) -> bool:
        # print(f'Entering GenericTypeHint._is_subhint_branch({self}, {branch})...')

        # If the unsubscripted type originating this generic is *NOT* a subclass
        # of the unsubscripted type originating that branch, this generic is
        # *NOT* a subhint of that branch. In this case, return false.
        if not issubclass(self._origin, branch._origin):
            # print(f'{self._origin} not subclass of {branch._origin})!')
            return False
        # Else, the unsubscripted type originating this generic is a subclass
        # of the unsubscripted type originating that branch. Note, however, that
        # this does *NOT* imply this generic to be a subhint of that branch.
        # The issubclass() builtin ignores parametrizations and thus returns
        # false positives for parametrized generics: e.g.,
        #     >>> from typing import Generic, TypeVar
        #     >>> T = TypeVar('T')
        #     >>> class MuhGeneric(Generic[T]): pass
        #     >>> issubclass(MuhGeneric, MuhGeneric[int])
        #     True
        #
        # Clearly, the unsubscripted generic "MuhGeneric" is a superhint
        # (rather than a subhint) of the subscripted generic
        # "MuhGeneric[int]". Further introspection is needed.
        #
        # If that branch is unsubscripted, assume that branch to have been
        # subscripted by "Any". Since *ANY* child hint subscripting this hint is
        # necessarily a subhint of "Any", this hint is a subhint of that branch.
        # Return true immediately.
        #
        # Note that this this common edge case implicitly handles comparison of
        # this generic against an unsubscripted simple class encapsulated by the
        # "ClassTypeHint" wrapper. Why? Because the
        # "ClassTypeHint._is_args_ignorable" property unconditionally returns
        # "True". Ergo, "ClassTypeHint" need *NOT* be handled below.
        elif branch._is_args_ignorable:
            # print(f'is_subhint_branch({self}, {branch} [unsubscripted])')
            return True
        # Else, that branch is subscripted.
        #
        # If this generic and that subscripted branch are incommensurable (i.e.,
        # encapsulated by incomparable type hint wrappers *NOT* supporting
        # comparison between a generic and a possibly non-generic type hint),
        # this generic is *NOT* a subhint of that branch. In this case, return
        # false.
        elif not isinstance(branch, _IS_SUBHINT_TYPES_COMMENSURABLE):
            # print(f'{branch.__class__} not in {_IS_SUBHINT_TYPES_COMMENSURABLE})!')
            return False
        # Else, this generic and that subscripted branch are commensurable
        # (i.e., encapsulated by comparable type hint wrappers supporting
        # comparison between a generic and a possibly non-generic type hint).
        #
        # By elimination, that branch *MUST* now be encapsulated by either:
        # * If that branch is also a PEP 484- or 585-compliant user-defined
        #   subscripted generic, "GenericTypeHint".
        # * Else, that branch *MUST* be a PEP 484- or 585-compliant subscripted
        #   non-generic (e.g., "list[int]", "collections.abc.Sized[str]").

        # Human-readable substring prefixing exception messages raised below.
        exception_prefix = f'{self} <= {branch} undecidable, as '

        # Tuple of the zero or more child hints transitively (i.e., *NOT*
        # directly) subscripting this generic *WITH RESPECT TO THAT BRANCH*.
        # This tuple generalizes the original tuple of child type hints directly
        # (i.e., *NOT* transitively) subscripting this generic. A generic is
        # transitively subscripted by the tuple of all child hints directly
        # subscripting this generic and all pseudo-superclasses of this generic.
        #
        # Deciding this tuple requires a highly non-trivial algorithm
        # performing a recursive depth-first search (DFS) over the
        # pseudo-superclass hierarchy implied by this generic. Doing so greedily
        # replaces in the original tuple as many abstract PEP 484-compliant type
        # variables (i.e., "typing.TypeVar" objects) as there are concrete child
        # hints directly subscripting this generic. Doing so effectively
        # "bubbles up" these concrete children up the class hierarchy into the
        # "empty placeholders" established by the type variables transitively
        # subscripting all pseudo-superclasses of this generic.  # <-- lolwat
        #
        # Note that this getter is memoized for efficiency and thus
        # intentionally accepts *NO* keyword parameters. It is what it is.
        self_args_full = get_hint_pep484585_generic_args_full(
            # This generic.
            self._hint,

            # The target pseudo-superclass of this generic to be searched for in
            # the pseudo-superclass hierarchy implied by this generic.
            branch._hint,  # pyright: ignore

            # Exception class to be raised in the event of fatal error.
            BeartypeDoorIsSubhintException,

            # Human-readable substring prefixing raised exception messages.
            exception_prefix,
        )
        # print(f'Found self {self} full child hints: {self_args_full}')

        #FIXME: Unit test us up, please.
        # If that branch is a subscripted non-generic (e.g., "Sequence[int]")...
        if isinstance(branch, SubscriptedTypeHint):
            # print(f'Comparing against subscripted non-generic {branch}...')

            # If these two hints are subscripted by a differing number of child
            # hints, raise an exception. See the superclass method for details.
            #
            # Note that this should *NEVER* occur. Therefore, this will occur.
            if len(self_args_full) != len(branch._args_wrapped_tuple):  # pragma: no cover
                # Number of child hints subscripting that branch.
                branch_args_len = len(branch._args_wrapped_tuple)

                # Raise an exception embedding these numbers.
                raise BeartypeDoorIsSubhintException(
                    f'{exception_prefix}'
                    f'{self._hint} transitively subscripted by child hints '
                    f'{self_args_full} whose length differs from '
                    f'{branch._hint} directly subscripted by child hints '
                    f'{branch.args} '
                    f'(i.e., {len(self_args_full)} != {branch_args_len}).'
                )
            # Else, these two hints are subscripted by the same number of child
            # hints.

            # For the 0-based index of each child hint transitively subscripting
            # this generic and this hint...
            for self_arg_full_index, self_arg_full in enumerate(self_args_full):
                # Hint wrapper encapsulating this child hint transitively
                # subscripting this generic.
                self_child = TypeHint(self_arg_full)

                # Hint wrapper encapsulating the corresponding child hint
                # directly subscripting that non-generic branch.
                branch_child = branch._args_wrapped_tuple[self_arg_full_index]

                # If this child hint is *NOT* a subhint of that child hint, this
                # generic is *NOT* a subhint of that non-generic branch. In this
                # case, short-circuit by immediately returning false.
                if not self_child.is_subhint(branch_child):
                    # print(f'{self_child} <= {branch_child} == False!')
                    return False
                # Else, this child hint is a subhint of that child hint. In this
                # case, this generic *COULD* be a subhint of that non-generic
                # branch. Decide by continuing to the next pair of child hints.

                # print(f'{self_child} <= {branch_child} == True...')
            # Else, each child hint of this generic is a subhint of the
            # corresponding child hint of that non-generic branch. In this case,
            # this generic is a subhint of that non-generic branch.
        # Else, that branch is *NOT* a subscripted non-generic
        # (e.g., "Sequence[int]"). By the above validation, that branch is both
        # subscripted and commensurable with this generic. By elimination, that
        # branch *MUST* be a subscripted generic (e.g., "MuhGeneric[T]"). In
        # this case...
        else:
            # print(f'Comparing against subscripted generic {branch}...')

            # Validate that that branch is, in fact, a subscripted generic.
            assert isinstance(branch, GenericTypeHint), (
                f'{branch} not PEP 484 or 585 subscripted generic '
                f'(i.e., "beartype.door.GenericTypeHint" instance).')

            # Tuple of the zero or more child hints transitively subscripting
            # that generic branch. See above for further details.
            branch_args_full = get_hint_pep484585_generic_args_full(
                # That generic branch.
                branch._hint,

                # *NO* target pseudo-superclass of that generic branch. Instead,
                # completely decide the full tuple of child hints transitively
                # subscripting that generic branch.
                None,

                # Exception class to be raised in the event of fatal error.
                BeartypeDoorIsSubhintException,

                # Human-readable substring prefixing raised exception messages.
                exception_prefix,
            )
            # print(f'Found branch {branch} full child hints: {branch_args_full}')

            # If these two generics are transitively subscripted by a differing
            # number of child hints, raise an exception. See above for details.
            #
            # Note that this should *NEVER* occur. Therefore, this will occur.
            if len(self_args_full) != len(branch_args_full):  # pragma: no cover
                # Raise an exception embedding these numbers.
                raise BeartypeDoorIsSubhintException(
                    f'{exception_prefix}'
                    f'{self._hint} transitively subscripted by child hints '
                    f'{self_args_full} whose length differs from '
                    f'{branch._hint} transitively subscripted by child hints '
                    f'{branch_args_full} '
                    f'(i.e., {len(self_args_full)} != {len(branch_args_full)}).'
                )
            # Else, these two generics are transitively subscripted by the same
            # number of child hints.

            # For the 0-based index of each child hint transitively subscripting
            # this generic and this hint...
            for self_arg_full_index, self_arg_full in enumerate(self_args_full):
                # Hint wrapper encapsulating this child hint transitively
                # subscripting this generic.
                self_child = TypeHint(self_arg_full)

                # Hint wrapper encapsulating the corresponding child hint
                # directly subscripting that generic branch.
                branch_child = TypeHint(branch_args_full[self_arg_full_index])
                # print(f'{self_child} <= {branch_child}?')

                # If this child hint is *NOT* a subhint of that child hint, this
                # generic is *NOT* a subhint of that generic branch. In this
                # case, short-circuit by immediately returning false.
                if not self_child.is_subhint(branch_child):
                    # print(f'{self_child} <= {branch_child} == False!')
                    return False
                # Else, this child hint is a subhint of that child hint. In this
                # case, this generic *COULD* be a subhint of that generic
                # branch. Decide by continuing to the next pair of child hints.

                # print(f'{self_child} <= {branch_child} == True...')
            # Else, each child hint of this generic is a subhint of the
            # corresponding child hint of that generic branch. In this case,
            # this generic is a subhint of that generic branch.

        # Return true! We have liftoff.
        return True

# ....................{ PRIVATE ~ constants                }....................
_IS_SUBHINT_TYPES_COMMENSURABLE = (
    # Generics are commensurable with other generics, clearly.
    GenericTypeHint,

    # Generics are also commensurable with subscripted type hints: e.g.,
    #     >>> from collections.abc import Sequence
    #     >>> class MuhSequence[T](Sequence[T]): pass
    #     >>> is_subhint(MuhSequence[int], Sequence[int])
    #     True
    SubscriptedTypeHint,

    # Generics are also commensurable with unsubscripted simple classes: e.g.,
    #     >>> from collections.abc import Sequence
    #     >>> class MuhSequence[T](Sequence[T]): pass
    #     >>> is_subhint(MuhSequence[T], Sequence)
    #     True
    ClassTypeHint,
)
'''
Tuple of all **commensurable generic types** (i.e., concrete :class:`.TypeHint`
subclasses such that the :meth:`GenericTypeHint._is_subhint_branch` method
returns :data:`False` if the passed ``branch`` parameter is *not* an instance of
a type in this tuple).

When passed a ``branch`` parameter that is *not* an instance of a type in this
tuple, the :meth:`GenericTypeHint._is_subhint_branch` method returns
:data:`False` due to that parameter being **incommensurable** (i.e.,
incomparable) with a :pep:`484`-compliant :class:`typing.Generic`.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **magic error substrings** (i.e., string constants intended to be
embedded in exception and warning messages or otherwise pertaining to exceptions
and warnings).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ STRINGS                            }....................
EXCEPTION_PLACEHOLDER = '$%ROOT_PITH_LABEL/~'
'''
Non-human-readable source substring to be globally replaced by a human-readable
target substring in the messages of memoized exceptions passed to the
:func:`reraise_exception` function.

This substring prefixes most exception messages raised by memoized callables,
including code generation factories memoized on passed PEP-compliant type hints
(e.g., the :mod:`beartype._check` and :mod:`beartype._decor` submodules). The
:func:`beartype._util.error.utilerrraise.reraise_exception_placeholder` function
then dynamically replaces this prefix of the message of the passed exception
with a human-readable synopsis of the current unmemoized exception context,
including the name of both the currently decorated callable *and* the currently
iterated parameter or return of that callable for aforementioned code generation
factories.

Usage
-----
This substring is typically hard-coded into non-human-readable exception
messages raised by low-level callables memoized with the
:func:`beartype._util.cache.utilcachecall.callable_cached` decorator. Why?
Memoization prohibits those callables from raising human-readable exception
messages. Why? Doing so would require those callables to accept fine-grained
parameters unique to each call to those callables, which those callables would
then dynamically format into human-readable exception messages raised by those
callables. The standard example would be a ``exception_prefix`` parameter
labelling the human-readable category of type hint being inspected by the
current call (e.g., ``@beartyped muh_func() parameter "muh_param" PEP type hint
"List[int]"`` for a ``List[int]`` type hint on the `muh_param` parameter of a
``muh_func()`` function decorated by the :func:`beartype.beartype` decorator).
Since the whole point of memoization is to cache callable results between calls,
any callable accepting any fine-grained parameter unique to each call to that
callable is effectively *not* memoizable in any meaningful sense of the
adjective "memoizable." Ergo, memoized callables *cannot* raise human-readable
exception messages unique to each call to those callables.

This substring indirectly solves this issue by inverting the onus of human
readability. Rather than requiring memoized callables to raise human-readable
exception messages unique to each call to those callables (which we've shown
above to be pragmatically infeasible), memoized callables instead raise
non-human-readable exception messages containing this substring where they
instead would have contained the human-readable portions of their messages
unique to each call to those callables. This indirection renders exceptions
raised by memoized callables generic between calls and thus safely memoizable.

This indirection has the direct consequence, however, of shifting the onus of
human readability from those lower-level memoized callables onto higher-level
non-memoized callables -- which are then required to explicitly (in order):

#. Catch exceptions raised by those lower-level memoized callables.
#. Call the :func:`reraise_exception_placeholder` function with those
   exceptions and desired human-readable substrings. That function then:

   #. Replaces this magic substring hard-coded into those exception messages
      with those human-readable substring fragments.
   #. Reraises the original exceptions in a manner preserving their original
      tracebacks.

Unsurprisingly, as with most inversion of control schemes, this approach is
non-intuitive. Surprisingly, however, the resulting code is actually *more*
elegant than the standard approach of raising human-readable exceptions from
low-level callables. Why? Because the standard approach percolates
human-readable substring fragments from the higher-level callables defining
those fragments to the lower-level callables raising exception messages
containing those fragments. The indirect approach avoids percolation, thus
streamlining the implementations of all callables involved. Phew!
'''


EXCEPTION_PREFIX_DEFAULT = f'{EXCEPTION_PLACEHOLDER}default '
'''
Non-human-readable source substring to be globally replaced by a human-readable
target substring in the messages of memoized exceptions passed to the
:func:`reraise_exception` function caused by violations raised when
type-checking the default values of optional parameters for
:func:`beartype.beartype`-decorated callables.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide :pep:`589`-compliant **typed dictionary generics** (i.e., user-defined
types subclassing both the :pep:`589`-compliant :obj:`typing.TypedDict`
pseudo-superclass *and* the :pep:`484`-compliant :obj:`typing.Generic`
superclass, exercising edge cases in unit tests requiring non-trivial generics).
'''

# ....................{ IMPORTS                            }....................
# Defer fixture-specific imports.
from beartype.typing import (
    Generic,
    TypedDict,
)
from beartype_test.a00_unit.data.pep.data_pep484 import T

# ....................{ PEP 484 ~ usable : T               }....................
# Generics that are actually instantiable and usable as valid objects.

class Pep589484TypedDictT(TypedDict, Generic[T]):
    '''
    :pep:`589`-compliant **typed dictionary generic** (i.e., user-defined
    type subclassing both the :pep:`589`-compliant :obj:`typing.TypedDict`
    pseudo-superclass *and* the :pep:`484`-compliant :obj:`typing.Generic`
    superclass).
    '''

    # ....................{ FIELDS                         }....................
    key: T
    '''
    Arbitrary :pep:`589`-compliant typed dictionary key annotated by the same
    :pep:`484`-compliant type variable parametrizing this typed dictionary.
    '''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Decidedly Object-Oriented Runtime-checking (DOOR) annotated type hint
classes** (i.e., :class:`beartype.door.TypeHint` subclasses implementing support
for :pep:`593`-compliant :attr:`typing.Annotated` type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.door._cls.doorsuper import TypeHint
from beartype._data.typing.datatypingport import Hint
from beartype._util.hint.pep.proposal.pep593 import (
    get_hint_pep593_metadata,
    get_hint_pep593_metahint,
)
from contextlib import suppress

# ....................{ SUBCLASSES                         }....................
class AnnotatedTypeHint(TypeHint):
    '''
    **Annotated type hint wrapper** (i.e., high-level object encapsulating a
    low-level :pep:`593`-compliant :attr:`typing.Annotated` type hint).

    Attributes
    ----------
    _metadata : tuple[object]
        **Metadata** (i.e., tuple of zero or more arbitrary low-level
        caller-defined objects annotating this :attr:`typing.Annotated` type
        hint, equivalent to all remaining arguments subscripting this hint).
    _metahint_wrapper : TypeHint
        **Metahint wrapper** (i.e., :class:`TypeHint` instance wrapping the
        child type hint annotated by this parent :attr:`typing.Annotated` type
        hint, equivalent to the first argument subscripting this hint).
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, hint: Hint) -> None:

        # Initialize our superclass.
        super().__init__(hint)

        # Tuple of the zero or more arbitrary caller-defined arguments following
        # the first argument subscripting this hint.
        self._metadata = get_hint_pep593_metadata(hint)

        # Wrapper wrapping the first argument subscripting this hint.
        self._metahint_wrapper = TypeHint(get_hint_pep593_metahint(hint))

    # ..................{ PRIVATE ~ properties               }..................
    @property
    def _is_args_ignorable(self) -> bool:
        # since Annotated[] must be used with at least two arguments, we are
        # never just the origin of the metahint
        return False

    # ..................{ PRIVATE ~ testers                  }..................
    def _is_equal(self, other: TypeHint) -> bool:

        return (
            isinstance(other, AnnotatedTypeHint)
            and self._metahint_wrapper == other._metahint_wrapper
            and self._metadata == other._metadata
        )


    def _is_subhint_branch(self, branch: TypeHint) -> bool:

        # If the other type is not annotated, we ignore annotations on this
        # one and just check that the metahint is a subhint of the other.
        # e.g. Annotated[t.List[int], 'meta'] <= List[int]
        if not isinstance(branch, AnnotatedTypeHint):
            return self._metahint_wrapper.is_subhint(branch)

        # Else, that hint is a "typing.Annotated[...]" type hint. If either...
        if (
            # The child type hint annotated by this parent hint does not subhint
            # the child type hint annotated by that parent hint *OR*...
            self._metahint_wrapper > branch._metahint_wrapper or
            # These hints are annotated by a differing number of objects...
            len(self._metadata) != len(branch._metadata)
        ):
            # This hint *CANNOT* be a subhint of that hint. Return false.
            return False

        # Attempt to...
        #
        # Note that the following iteration performs equality comparisons on
        # arbitrary caller-defined objects. Since these comparisons may raise
        # arbitrary caller-defined exceptions, we silently squelch any such
        # exceptions that arise by returning false below instead.
        with suppress(Exception):
            # Return true only if these hints are annotated by equivalent
            # objects. We avoid testing for a subhint relation here (e.g., with
            # the "<=" operator), as arbitrary caller-defined objects are *MUCH*
            # more likely to define a relevant equality comparison than a
            # relevant less-than-or-equal-to comparison.
            return self._metadata == branch._metadata

        # Else, one or more objects annotating these hints are incomparable. So,
        # this hint *CANNOT* be a subhint of that hint. Return false.
        return False  # pragma: no cover



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **abstract syntax tree (AST) functionality** (i.e., low-level callables
and classes instrumenting well-typed third-party modules with runtime
type-checking dynamically generated by the :func:`beartype.beartype` decorator).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: [PEP 484] Additionally define generator transformers. The idea here is
#that @beartype *CAN* actually automatically type-check generator yields, sends,
#and returns at runtime. How? By automatically injecting appropriate
#die_if_unbearable() calls type-checking the values to be yielded, sent, and
#returned against the appropriate type hints of the current generator factory
#*BEFORE* yielding, sending, and returning those values. Shockingly, typeguard
#already does this -- which is all manner of impressive. See the
#TypeguardTransformer._use_memo() context manager for working code. Wow!
#    https://github.com/agronholm/typeguard/blob/master/src/typeguard/_transformer.py
#
#Note, however, that doing so is extremely complicated in the modern Python era.
#Why? PEP 649. Unquoted forward references in type hints are now resolved
#dynamically at runtime via non-trivial machinery that, ultimately, calls
#__annotate__() dunder methods declared on callables. This means that type hints
#are no longer safely copyable at early AST transformation time, which in turn
#means that AST transformers *CANNOT* safely copy type hints annotating
#callables into parameters passed to die_if_unbearable() calls. Is there no hope
#then? There is hope, actually -- and fairly trivial, too. How? Just reference
#rather than copy nested generator type hints. Since the @beartype decorator
#guarantees nested generator type hints to be resolved by the time these
#die_if_unbearable() calls are actually performed, referencing these type hints
#suffices to preserve sanity: e.g.,
#    from collections.abc import Iterable
#    def muh_generator_func(muh_arg: str) -> Iterable[str]:
#        die_in_unbearable(
#            muh_arg, muh_generator_func.__annotations__['muh_arg])
#        yield muh_arg
#
#Of course, even the convoluted pseudocode above still fails to suffice. Why?
#Because the "__annotations__" dunder dictionary is no longer safely accessible.
#Doing so raises "NameError" exceptions for type hints subscripted by one or
#more unquoted forward references. Instead, we'll need to call some sort of PEP
#649-compliant private beartype utility function (which doesn't even exist,
#currently) to efficiently retrieve specific type hints.
#
#Super-annoying, but also super-feasible. Still, nobody cares at the moment.
#Let's let this one slip onto the back burner for the moment, eh?

#FIXME: [SPEED] Consider generalizing the BeartypeNodeTransformer.__new__()
#class method to internally cache and return "BeartypeNodeTransformer" instances
#depending on the passed "conf" parameter. In general, most codebases
#will only leverage a single @beartype configuration (if any @beartype
#configuration at all); ergo, caching improves everything by enabling us to
#reuse the same "BeartypeNodeTransformer" instance for every hooked module.
#Score @beartype! See the BeartypeConf.__new__() method for relevant logic. \o/
#FIXME: Oh, wait. We probably do *NOT* want to cache -- at least, not without
#defining a comparable reinit() method as we do for "BeartypeDecorMeta". After
#retrieving a cached "BeartypeNodeTransformer" instance, we'll need to
#immediately call BeartypeNodeTransformer.reinit() to reinitialize that
#instance.
#
#This is all feasible, of course -- but let's just roll with the naive
#implementation for now, please.

#FIXME: [PEP 675] *OMG.* See also the third-party "executing" Python package:
#    https://github.com/alexmojaki/executing
#
#IPython itself internally leverages "executing" via "stack_data" (i.e., a
#slightly higher-level third-party Python package that internally leverages
#"executing") to syntax-highlight the currently executing AST node. Indeed,
#"executing" sports an intense test suite (much like ours) effectively
#guaranteeing a one-to-one mapping between stack frames and AST nodes.
#
#So, what's the Big Idea here? The Big Idea here is that @beartype can
#internally (...possibly only optionally, but possibly mandatorily) leverage
#"executing" to begin performing full-blown static type-checking at runtime --
#especially of mission critical type hints like "typing.LiteralString" which can
#*ONLY* be type-checked via static analysis. :o
#
#So, what's the Little Idea here? The Little Idea here is that @beartype can
#generate type-checking wrappers that type-check parameters or returns annotated
#by "typing.LiteralString" by calling an internal private utility function --
#say, "_die_unless_literalstring(func: Callable, arg_name: str) -> None" -- with
#"func" as the current type-checking wrapper and "arg_name" as either the name
#of that parameter or "return". The _die_unless_literalstring() raiser then:
#* Dynamically searches up the call stack for the stack frame encapsulating an
#  external call to the passed "func" callable.
#* Passes that stack frame to the "executing" package.
#* "executing" then returns the AST node corresponding to that stack frame.
#* Introspects that node for the passed parameter whose name is "arg_name".
#* Raises an exception unless the value of that parameter is an AST node
#  corresponding to a string literal.
#
#Of course, that won't necessarily be fast -- but it will be accurate. Since
#security trumps speed, speed is significantly less of a concern insofar as
#"typing.LiteralString" is concerned. Of course, we should also employ
#significant caching... if we even can.
#FIXME: Actually, while demonstrably awesome, even the above fails to suffice to
#to statically type-check "typing.LiteralString". We failed to fully read PEP
#675, which contains a section on inference. In the worst case, nothing less
#than a complete graph of the entire app and all transitive dependencies thereof
#suffices to decide whether a parameter satisfies "typing.LiteralString".
#
#Thankfully, the above idea generalizes from "typing.LiteralString" to other
#fascinating topics as well. Indeed, given sufficient caching, one could begin
#to internally generate and cache a mypy-like graph network whose nodes are
#typed attributes and whose edges are relations between those typed attributes.



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

# ....................{ TESTS                              }....................
'''
Beartype :pep:`544`-compliant **type hint reduction** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._check.convert._reduce._pep.redpep544` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ reducers                   }....................
def test_reduce_hint_pep484_generic_io_to_pep544_protocol() -> None:
    '''
    Test the
    :func:`beartype._check.convert._reduce._pep.redpep544.reduce_hint_pep484_generic_io_to_pep544_protocol`
    reducer.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep544Exception
    from beartype.typing import Union
    from beartype._check.convert._reduce._pep.redpep544 import (
        reduce_hint_pep484_generic_io_to_pep544_protocol)
    from beartype._data.cls.datacls import TYPES_PEP484_GENERIC_IO
    from beartype._util.hint.pep.proposal.pep593 import is_hint_pep593
    from pytest import raises
    from typing import Protocol

    # ....................{ PASS                           }....................
    # For each PEP 484-compliant "typing" IO generic base class...
    for pep484_generic_io in TYPES_PEP484_GENERIC_IO:
        # Equivalent protocol reduced from this generic.
        pep544_protocol_io = (
            reduce_hint_pep484_generic_io_to_pep544_protocol(
                pep484_generic_io, ''))

        # Assert this protocol is either...
        assert (
            # A PEP 593-compliant type metahint generalizing a protocol
            # *OR*...
            is_hint_pep593(pep544_protocol_io) or
            # A PEP 544-compliant protocol.
            issubclass(pep544_protocol_io, Protocol)
        )

    # ....................{ FAIL                           }....................
    # Assert this function rejects standard type hints in either case.
    with raises(BeartypeDecorHintPep544Exception):
        reduce_hint_pep484_generic_io_to_pep544_protocol(Union[int, str], '')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **import hook factories** (i.e., low-level utility functions creating
and returning objects of interest to higher-level import hook functions).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.claw._package.clawpkgenum import BeartypeClawCoverage
from beartype.roar import (
    BeartypeClawDecorWarning,
    BeartypeClawHookException,
)
from beartype.typing import Optional
from beartype._conf.confmain import BeartypeConf
from beartype._conf.conftest import die_unless_conf
from beartype._data.typing.datatyping import IterableStrs
from beartype._util.text.utiltextidentifier import die_unless_identifier
from collections.abc import Iterable as IterableABC

# ....................{ PRIVATE ~ factories                }....................
#FIXME: Unit test us up, please.
def make_conf_hookable(conf: BeartypeConf) -> BeartypeConf:
    '''
    New **hookable beartype configuration** (i.e., beartype configuration
    suitable for use in import hooks, sanitized from the passed beartype
    configuration which is typically unsuitable for use in import hooks).

    This getter creates and returns a new configuration permuted from the passed
    configuration, forcefully enabling these parameters required by import
    hooks:

    * :attr:`beartype.BeartypeConf.warning_cls_on_decorator_exception`
      to the :class:`beartype.roar.BeartypeClawDecorWarning` warning category.
      Doing so instructs the :func:`beartype.beartype` decorator to emit
      non-fatal warnings rather than raise fatal exceptions at decoration time
      when implicitly decorating callables and classes defined by modules hooked
      by our import hooks, substantially improving the robustness and usability
      of those hooks.

    Returns
    -------
    Optional[Iterable[str]]
        Iterable of the fully-qualified names of one or more packages to be
        either hooked or unhooked by the parent call.

    Raises
    ------
    BeartypeConfException
        If the passed ``conf`` parameter is *not* a beartype configuration
        (i.e., :class:`BeartypeConf` instance).

    See Also
    --------
    :func:`.hook_packages`
        Further details.
    '''

    # If this configuration is invalid, raise an exception.
    die_unless_conf(conf=conf, exception_cls=BeartypeClawHookException)
    # Else, this configuration is valid.

    # If the caller did *NOT* explicitly set the
    # "warning_cls_on_decorator_exception" configuration parameter governing the
    # reduction of fatal exceptions to non-fatal warnings at @beartype
    # decoration-time...
    if not conf._is_warning_cls_on_decorator_exception_set:
        # Keyword dictionary with which to instantiate a new configuration
        # reducing fatal exceptions to non-fatal warnings of a warning category
        # specific to beartype import hooks.
        conf_kwargs = conf.kwargs.copy()
        conf_kwargs['warning_cls_on_decorator_exception'] = (
            BeartypeClawDecorWarning)

        # Replace this configuration with this new configuration.
        conf = BeartypeConf(**conf_kwargs)  # type: ignore[arg-type]
    # Else, this caller already explicitly set the
    # "warning_cls_on_decorator_exception" configuration parameter governing the
    # reduction of fatal exceptions to non-fatal warnings at @beartype
    # decoration-time. In this case, preserve this user-defined reduction as is.

    # Return this possibly new configuration.
    return conf


#FIXME: Unit test us up, please.
def make_package_names_from_args(
    # Keyword-only arguments.
    *,

    # Mandatory keyword-only arguments.
    claw_coverage: BeartypeClawCoverage,
    conf: BeartypeConf,

    # Optional keyword-only arguments.
    package_name: Optional[str] = None,
    package_names: Optional[IterableStrs] = None,
) -> Optional[IterableStrs]:
    '''
    Validate all parameters passed by the caller to the current call of the
    parent :func:`beartype.claw._package.clawpkgmain.hook_packages` or
    :func:`beartype.claw._package.clawpkgmain.unhook_packages` function.

    Parameters
    ----------
    All keyword-only parameters accepted by the current call of the
    parent :func:`beartype.claw._package.clawpkgmain.hook_packages` or
    :func:`beartype.claw._package.clawpkgmain.unhook_packages` function.

    Returns
    -------
    Optional[Iterable[str]]
        Iterable of the fully-qualified names of one or more packages to be
        either hooked or unhooked by the parent call.

    Raises
    ------
    BeartypeClawHookException
        If the passed ``package_names`` parameter is either:

        * Neither a string nor an iterable (i.e., fails to satisfy the
          :class:`collections.abc.Iterable` protocol).
        * An empty string or iterable.
        * A non-empty string that is *not* a valid **package name** (i.e.,
          ``"."``-delimited concatenation of valid Python identifiers).
        * A non-empty iterable containing at least one item that is either:

          * *Not* a string.
          * The empty string.
          * A non-empty string that is *not* a valid **package name** (i.e.,
            ``"."``-delimited concatenation of valid Python identifiers).
    '''
    assert isinstance(conf, BeartypeConf), f'{repr(conf)} not configuration.'
    assert isinstance(claw_coverage, BeartypeClawCoverage), (
        f'{repr(claw_coverage)} not beartype claw coverage.')

    # If the caller requested all-packages coverage...
    if claw_coverage is BeartypeClawCoverage.PACKAGES_ALL:
        # If the caller improperly passed a package name despite requesting
        # all-packages coverage, raise an exception.
        if package_name is not None:
            raise BeartypeClawHookException(
                f'Coverage {repr(BeartypeClawCoverage.PACKAGES_ALL)} '
                f'but package name {repr(package_name)} passed.'
            )
        # Else, the caller properly passed *NO* package name.
        #
        # If the caller improperly passed multiple package names despite
        # requesting all-packages coverage, raise an exception.
        elif package_names is not None:
            raise BeartypeClawHookException(
                f'Coverage {repr(BeartypeClawCoverage.PACKAGES_ALL)} '
                f'but package names {repr(package_names)} passed.'
            )
        # Else, the caller properly passed *NO* package names.
    # Else, the caller did *NOT* request all-packages coverage. In this case,
    # the caller requested coverage over only a subset of packages.
    else:
        # If the caller requested mono-package coverage...
        if claw_coverage is BeartypeClawCoverage.PACKAGES_ONE:
            # If the caller improperly passed *NO* package name despite
            # requesting mono-package coverage, raise an exception.
            if package_name is None:
                raise BeartypeClawHookException(
                    f'beartype_package() '
                    f'package name {repr(package_name)} invalid.'
                )
            # Else, the caller properly passed a package name.

            # Wrap this package name in a 1-tuple containing only this name.
            # Doing so unifies logic below.
            package_names = (package_name,)
        # Else, the caller requested multi-package coverage.
        # elif coverage is BeartypeClawCoverage.PACKAGES_MANY:

        # If this package names is *NOT* iterable, raise an exception.
        if not isinstance(package_names, IterableABC):
            raise BeartypeClawHookException(
                f'beartype_packages() '
                f'package names {repr(package_name)} not iterable.'
            )
        # Else, this package names is iterable.
        #
        # If *NO* package names were passed, raise an exception.
        elif not package_names:
            raise BeartypeClawHookException(
                'beartype_packages() package names empty.')
        # Else, one or more package names were passed.

        # For each such package name...
        for package_name in package_names:
            # If this package name is *NOT* a string, raise an exception.
            if not isinstance(package_name, str):
                raise BeartypeClawHookException(
                    f'Package name {repr(package_name)} not string.')
            # Else, this package name is a string.
            #
            # If this package name is *NOT* a valid Python identifier, raise an
            # exception.
            else:
                die_unless_identifier(
                    text=package_name,
                    exception_cls=BeartypeClawHookException,
                    exception_prefix='Package name ',
                )
            # Else, this package name is a valid Python identifier.

    # Return the iterable of the fully-qualified names of one or more packages
    # to be either hooked or unhooked by the parent call.
    return package_names



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484` and :pep:`585` **callable type hint utility unit
tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep484585.pep484585callable` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_get_hint_pep484585_callable_params_and_return(hints_pep_meta) -> None:
    '''
    Test both the ``get_hint_pep484585_callable_params`` and
    ``get_hint_pep484585_callable_return`` getters declared by the
    :mod:`beartype._util.hint.pep.proposal.pep484585.pep484585callable`
    submodule.

    Since these getters are inextricably interrelated, this unit test exercises
    both within the same test to satisfy Don't Repeat Yourself (DRY).

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of PEP-compliant type hint metadata describing sample PEP-compliant
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep484585Exception
    from beartype._data.hint.sign.datahintsigns import HintSignCallable
    from beartype._util.hint.pep.proposal.pep484585.pep484585callable import (
        get_hint_pep484585_callable_params,
        get_hint_pep484585_callable_return,
    )
    from beartype_test.a00_unit.data.hint.data_hint import NOT_HINTS_PEP
    from pytest import raises
    from typing import (
        Any,
        Concatenate,
        ParamSpec,

        # Intentionally import the callable type hint factory from "typing"
        # rather than "beartype.typing" to guarantee PEP 484-compliance.
        Callable as Pep484Callable,
    )

    # Intentionally import the callable type hint factory from "collections.abc"
    # rather than "beartype.typing" to guarantee PEP 585-compliance.
    from collections.abc import Callable as Pep585Callable

    # ..................{ GENERIC                            }..................
    # General-purpose logic generically exercising these getters against all
    # globally defined type hints.

    # Assert these getters...
    for hint_pep_meta in hints_pep_meta:
        # Return zero or more arguments for PEP-compliant callable type hints.
        if hint_pep_meta.pep_sign is HintSignCallable:
            hint_callable_params = get_hint_pep484585_callable_params(
                hint_pep_meta.hint)
            hint_callable_return = get_hint_pep484585_callable_return(
                hint_pep_meta.hint)

            assert isinstance(hint_callable_params, tuple)
            assert hint_callable_return is not None
        # Raise an exception for concrete PEP-compliant type hints *NOT*
        # defined by the "typing" module.
        else:
            with raises(BeartypeDecorHintPep484585Exception):
                get_hint_pep484585_callable_params(hint_pep_meta.hint)
            with raises(BeartypeDecorHintPep484585Exception):
                get_hint_pep484585_callable_return(hint_pep_meta.hint)

    # Assert these getters raise the expected exception for non-PEP-compliant
    # type hints.
    for not_hint_pep in NOT_HINTS_PEP:
        with raises(BeartypeDecorHintPep484585Exception):
            get_hint_pep484585_callable_params(not_hint_pep)
        with raises(BeartypeDecorHintPep484585Exception):
            get_hint_pep484585_callable_return(not_hint_pep)

    # ..................{ LOCALS                             }..................
    # Arbitrary PEP 612-compliant child hints.
    P = ParamSpec('P')
    str_plus_P = Concatenate[str, P]

    # ..................{ CASES                              }..................
    # List of 3-tuples "(callable_hint, callable_hint_params,
    # callable_hint_return)", where:
    # * "callable_hint" is a PEP-compliant callable type hint to be tested.
    # * "callable_hint_params" is the parameters type hint subscripting that
    #   callable type hint.
    # * "callable_hint_return" is the return type hint subscripting that
    #   callable type hint.
    CALLABLE_HINT_PARAMS_RETURN_CASES = [
        # ..................{ PEP ~ 484                      }..................
        # PEP 484-compliant callable type hints.
        (Pep484Callable[[], Any], (), Any),
        (Pep484Callable[[int], bool], (int,), bool),
        (Pep484Callable[[int, bool], float], (int, bool), float),
        (Pep484Callable[..., bytes], Ellipsis, bytes),

        # ..................{ PEP ~ 585                      }..................
        # Extend this list with PEP 585-compliant callable type hints.
        (Pep585Callable[[], Any], (), Any),
        (Pep585Callable[[int], bool], (int,), bool),
        (Pep585Callable[[int, bool], float], (int, bool), float),
        (Pep585Callable[..., bytes], Ellipsis, bytes),

        # Note this edge case is intentionally *NOT* tested above as a
        # PEP 484-compliant callable type hint, as the "typing.Callable"
        # factory refuses to accept the empty tuple here: e.g.,
        #     >>> from typing import Callable
        #     >>> Callable[[()], str]
        #     TypeError: Callable[[arg, ...], result]: each arg must be a
        #     type. Got ().
        (Pep585Callable[[()], str], (), str),

        # ..................{ PEP ~ 612                      }..................
        # Extend this list with PEP 612-compliant callable type hints.
        (Pep585Callable[P, Any], P, Any),
        (Pep585Callable[str_plus_P, int], str_plus_P, int),
    ]

    # ..................{ PEP                                }..................
    # For each callable type hint defined above...
    for hint, hint_params, hint_return in CALLABLE_HINT_PARAMS_RETURN_CASES:
        # Parameters type hint returned by this getter for this hint.
        hint_params_actual = get_hint_pep484585_callable_params(hint)

        # If the parameters type hint subscripting this callable type hint is a
        # tuple, assert this tuple to be equal but *NOT* identical to this
        # actual tuple.
        if isinstance(hint_params, tuple):
            assert hint_params_actual == hint_params
        # Else, the parameters type hint subscripting this callable type hint is
        # a non-tuple. In this case, assert this non-tuple to be identical to
        # this actual non-tuple.
        else:
            assert hint_params_actual is hint_params

        # Assert this getter returns the expected return.
        assert get_hint_pep484585_callable_return(hint) is hint_return



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **sentinel singletons** (i.e., objects of arbitrary placeholder
value commonly required throughout this codebase, reducing space and time
consumption by preallocating widely used sentinel objects).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ CLASSES                            }....................
class Iota(object):
    '''
    **Iota** (i.e., object minimizing space consumption by guaranteeably
    containing *no* attributes).
    '''

    __slots__ = ()


    def __repr__(self) -> str:
        '''
        Machine-readable representation of this iota.
        '''

        # Return the fully-qualified name of the sentinel placeholder defined
        # below. Since this is the *ONLY* meaningful instance of this type
        # instantiated throughout the codebase, this reduction improves the
        # readability of debugging messages and logging.
        return 'beartype._data.kind.datakindiota.SENTINEL'

# ....................{ CONSTANTS                          }....................
SENTINEL = Iota()
'''
**Sentinel singleton** (i.e., object of arbitrary placeholder value).

This object is internally leveraged by various utility functions to identify
erroneous and edge-case input (e.g., iterables of insufficient length).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **callable parameter getter utilities** (i.e., low-level functions
introspecting metadata on various kinds of parameters accepted by arbitrary
callables).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilCallableException
from beartype.typing import (
    Optional,
    Tuple,
)
from beartype._data.typing.datatyping import (
    Codeobjable,
    TypeException,
)
from beartype._util.func.arg.utilfuncargiter import (
    ARG_META_INDEX_KIND,
    ArgKind,
    ArgMeta,
    iter_func_args,
)
from beartype._util.func.arg.utilfuncarglen import get_func_args_len
from beartype._util.func.utilfunccodeobj import get_func_codeobj

# ....................{ GETTERS ~ meta                     }....................
#FIXME: [SPEED] Consider refactoring this O(n) iteration into an O(1) lookup.
#Although extremely non-trivial, doing so *SHOULD* be feasible based on the
#current internal implementation of the iter_func_args() generator
#comprehension. Basically, we'll need to:
#* Augment get_func_args_lens() to additionally cache and return *ALL* of
#  the remaining "sub-lengths" internally computed by iter_func_args().
#  *ALL*. And there are quite a few, indeed.
#* Call get_func_args_lens() here.
#* Get the name of the variadic positional parameter by simply indexing:
#      func_args_len = get_func_args_len(*args, **kwargs)
#      func_codeobj.co_varnames[func_args_len + hur_der_hur_something]
#  ...where "hur_der_hur_something" is the index of the variadic positional
#  parameter in the "co_varnames" list, computed by adding all requisite
#  "sub-lengths" together. We shrug.
#FIXME: Unit test us up, please.
def get_func_arg_meta_variadic_positional_or_none(
    *args, **kwargs) -> Optional[ArgMeta]:
    '''
    Callable parameter metadata describing the variadic positional parameter
    accepted by the passed pure-Python callable if that callable accepts a
    variadic positional parameter *or* :data:`None` otherwise (i.e., if that
    callable accepts no variadic positional parameter).

    Caveats
    -------
    **This getter exhibits worst-case linear-time complexity** :math:`O(n)`
    **for** :math:`O(n)` **the total number of parameters accepted by that
    callable** due to unavoidably performing a linear search for a variadic
    positional parameter with this name is this callable's parameter list. This
    getter should thus be called sparingly.

    Parameters
    ----------
    All remaining parameters are passed as is to the lower-level
    :func:`beartype._util.func.utilfunccodeobj.get_func_codeobj` getter.

    Returns
    -------
    Optional[ArgMeta]
        Either:

        * If that callable accepts a variadic positional parameter, callable
          parameter metadata describing that parameter.
        * Else, :data:`None`.

    Raises
    ------
    exception_cls
         If the passed callable is *not* pure-Python.
    '''

    # Avoid circular import dependencies.
    from beartype._util.func.arg.utilfuncargtest import (
        is_func_arg_variadic_positional)

    # If the passed callable accepts *NO* variadic positional parameter,
    # silently reduce to a noop.
    if not is_func_arg_variadic_positional(*args, **kwargs):
        return None
    # Else, that callable accepts a variadic positional parameter.

    # For metadata describing each parameter accepted by that callable...
    for arg_meta in iter_func_args(*args, **kwargs):
        # If this is a variadic positional parameter, return this metadata.
        if arg_meta[ARG_META_INDEX_KIND] is ArgKind.VARIADIC_POSITIONAL:
            return arg_meta
        # Else, this is *NOT* a variadic positional parameter. In this case,
        # silently continue to the next parameter.
    # Else, that callable accepts *NO* variadic positional parameter. But, by
    # the above validation, that callable accepts a variadic positional
    # parameter! We are now off the deep end.

    # Raise an exception as a last-ditch fallback.
    #
    # Note that this should *NEVER* occur. Naturally, this just occurred.
    raise _BeartypeUtilCallableException(  # pragma: no cover
        f'Callable inexplicably accepts no variadic positional parameter '
        f'despite claiming to do so:\n\t{repr(args)}\n\t{repr(kwargs)}'
    )


#FIXME: Unit test us up, please.
def get_func_arg_meta_variadic_keyword_or_none(
    *args, **kwargs) -> Optional[ArgMeta]:
    '''
    Callable parameter metadata describing the variadic keyword parameter
    accepted by the passed pure-Python callable if that callable accepts a
    variadic keyword parameter *or* :data:`None` otherwise (i.e., if that
    callable accepts no variadic keyword parameter).

    Caveats
    -------
    **This getter exhibits worst-case linear-time complexity** :math:`O(n)`
    **for** :math:`O(n)` **the total number of parameters accepted by that
    callable** due to unavoidably performing a linear search for a variadic
    keyword parameter with this name is this callable's parameter list. This
    getter should thus be called sparingly.

    Parameters
    ----------
    All remaining parameters are passed as is to the lower-level
    :func:`beartype._util.func.utilfunccodeobj.get_func_codeobj` getter.

    Returns
    -------
    Optional[ArgMeta]
        Either:

        * If that callable accepts a variadic keyword parameter, callable
          parameter metadata describing that parameter.
        * Else, :data:`None`.

    Raises
    ------
    exception_cls
         If the passed callable is *not* pure-Python.
    '''

    # Avoid circular import dependencies.
    from beartype._util.func.arg.utilfuncargtest import (
        is_func_arg_variadic_keyword)

    # If the passed callable accepts *NO* variadic keyword parameter,
    # silently reduce to a noop.
    if not is_func_arg_variadic_keyword(*args, **kwargs):
        return None
    # Else, that callable accepts a variadic keyword parameter.

    # For metadata describing each parameter accepted by that callable...
    for arg_meta in iter_func_args(*args, **kwargs):
        # If this is a variadic keyword parameter, return this metadata.
        if arg_meta[ARG_META_INDEX_KIND] is ArgKind.VARIADIC_KEYWORD:
            return arg_meta
        # Else, this is *NOT* a variadic keyword parameter. In this case,
        # silently continue to the next parameter.
    # Else, that callable accepts *NO* variadic keyword parameter. But, by
    # the above validation, that callable accepts a variadic keyword
    # parameter! We are now off the deep end.

    # Raise an exception as a last-ditch fallback.
    #
    # Note that this should *NEVER* occur. Naturally, this just occurred.
    raise _BeartypeUtilCallableException(  # pragma: no cover
        f'Callable inexplicably accepts no variadic keyword parameter '
        f'despite claiming to do so:\n\t{repr(args)}\n\t{repr(kwargs)}'
    )

# ....................{ GETTERS ~ name                     }....................
def get_func_arg_name_first_or_none(*args, **kwargs) -> Optional[str]:
    '''
    Name of the first parameter listed in the signature of the passed
    pure-Python callable if any *or* :data:`None` otherwise (i.e., if that
    callable accepts *no* parameters and is thus parameter-less).

    Parameters
    ----------
    All arguments are passed as is to the lower-level
    :func:`.get_func_arg_names` getter.

    Returns
    -------
    Optional[str]
        Either:

        * If that callable accepts one or more parameters, the name of the first
          parameter listed in the signature of that callable.
        * Else, :data:`None`.

    Raises
    ------
    exception_cls
         If that callable is *not* pure-Python.
    '''

    # Tuple of the names of all parameters accepted by the passed callable.
    func_arg_names = get_func_arg_names(*args, **kwargs)

    # Return either...
    return (
        # If that callable accepts one or more parameters, the name of the first
        # parameter accepted by that callable;
        func_arg_names[0]
        if func_arg_names else
        # Else, that callable accepts *NO* parameters. In this case, "None".
        None
    )


def get_func_arg_names(
    # Mandatory parameters.
    func: Codeobjable,

    # Optional parameters.
    is_omit_boundmethod_arg_first: bool = True,
    is_unwrap: bool = True,
    exception_cls: TypeException = _BeartypeUtilCallableException,
    exception_prefix: str = '',
) -> Tuple[str, ...]:
    '''
    Tuple of the names of all parameters accepted by the passed pure-Python
    callable.

    Ideally, this getter would have returned a set rather than a tuple. Sadly,
    sets *still* fail to guarantee insertion order despite comparable
    dictionaries doing so. A tuple is the next-best-thing. It is what it is.

    Caveats
    -------
    **Order is insignificant.** Ideally, this getter would return parameter
    names in the same order as the original callable signature declared those
    parameters. While feasible, doing so would require manual iteration over
    parameters exhibiting :math:`O(n)` time complexity. Instead, this getter
    performs a trivial slicing the ``co_varnames`` instance variable of the code
    object of that callable exhibiting :math:`O(1)` time complexity. Although
    extremely fast, doing so has the minor disadvantage of returning parameter
    names in an occasionally unexpected order. The sole exception to this rule
    is the first parameter name returned by this getter, which is guaranteed to
    be the first parameter declared by that callable.

    Parameters
    ----------
    func : Codeobjable
        Pure-Python callable, frame, or code object to be inspected.
    is_omit_boundmethod_arg_first : bool, optional
        :data:`True` only if this getter implicitly omits the first mandatory
        flexible parameter accepted by that callable if that callable is a
        C-based bound method descriptor encapsulating either an instance method
        bound to an instance of a class *or* a class method bound to a class.
        Defaults to :data:`True`, instructing this getter to transparently yield
        the *actual* high-level parameters accepted by this bound method
        descriptor (rather than the low-level parameters accepted by the unbound
        method encapsulated by this bound method descriptor). While the default
        behaviour is typically desirable, valid use cases for the non-default
        behaviour do exist (e.g., crudely detecting the kind of that based on
        whether the unbound method encapsulated by this bound method descriptor
        accepts a first parameter named ``cls`` or ``self``).

        The default behaviour enables:

        * This getter to transparently support bound method descriptors, which
          then enables...
        * The private :func:`beartype._decor._nontype.decornontype.beartype_pseudofunc`
          decorator to type-check the bound ``__call__()`` method descriptor
          encapsulating the unbound ``__call__()`` dunder method defined on the
          class of pseudo-callable objects, which then enables...
        * The public :func:`beartype.beartype` decorator to type-check
          pseudo-callable objects.

        How? In this case, the aforementioned ``beartype_pseudofunc``
        decorator wraps this bound method descriptor with a dynamically
        generated wrapper function that does *not* accept a ``self`` or ``cls``
        parameter, since a bound method does *not* accept a ``self`` or ``cls``
        parameter. After all, that's why bound methods exist; they implicitly
        pass the instance or class to which they are bound as the value of the
        ``self`` or ``cls`` parameter to the unbound method they encapsulate.
        However, the code object of a bound method descriptor is only an alias
        to the code object of the corresponding unbound method. Since the latter
        accepts a ``self`` parameter, so too does the former.

        The default behaviour resolves this internal discrepancy (contradiction)
        that arises between:

        * The code object of a bound method descriptor, which declares that
          callable object to accept a ``self`` parameter.
        * The real-world calling semantics of a bound method descriptor, which
          by definition accepts *no* ``self`` parameter.
    is_unwrap: bool, optional
        :data:`True` only if this getter implicitly calls the
        :func:`beartype._util.func.utilfuncwrap.unwrap_func_all` function.
        Defaults to :data:`True` for safety. See :func:`.get_func_codeobj` for
        further commentary.
    exception_cls : type, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :class:`._BeartypeUtilCallableException`.
    exception_prefix : str, optional
        Human-readable label prefixing the message of any exception raised in
        the event of a fatal error. Defaults to the empty string.

    Returns
    -------
    Tuple[str, ...]
        Tuple of the names of all parameters accepted by that callable.

    Raises
    ------
    exception_cls
         If the passed callable is *not* pure-Python.
    '''
    assert isinstance(is_omit_boundmethod_arg_first, bool), (
        f'{repr(is_omit_boundmethod_arg_first)} not boolean.')

    # Avoid circular import dependencies.
    from beartype._util.func.utilfunctest import is_func_boundmethod

    # Code object underlying the passed callable if any *OR* "None" otherwise
    # (i.e., that callable has *NO* code object).
    func_codeobj = get_func_codeobj(
        func=func,
        is_unwrap=is_unwrap,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )

    # Total number of parameters accepted by that callable.
    func_args_len = get_func_args_len(
        func=func_codeobj,
        is_unwrap=False,  # <-- "func" was already unwrapped above. I sigh.
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )

    # Tuple of the names of all parameters accepted by the passed callable.
    func_arg_names = func_codeobj.co_varnames[:func_args_len]  # <-- BOOM

    # If...
    if (
        # If that callable accepts one or more parameters *AND*...
        func_arg_names and
        # Omitting the first mandatory flexible parameter accepted by that
        # callable if that callable is a C-based bound method descriptor
        # encapsulating either an instance method bound to an instance of a
        # class or a class method bound to a class *AND*...
        is_omit_boundmethod_arg_first and
        # That callable is such a C-based bound method descriptor...
        is_func_boundmethod(func)
    ):
        # Omit the first parameter accepted by that bound method descriptor.
        #
        # Note that that callable is guaranteed to accept at least one
        # parameter, due to being a bound method descriptor.
        func_arg_names = func_arg_names[1:]

    # Return this tuple.
    return func_arg_names
